/******************************************************************************
 *                                                                            *
 *                                Copyright by                                *
 *                                                                            *
 *                              Azoteq (Pty) Ltd                              *
 *                          Republic of South Africa                          *
 *                                                                            *
 *                           Tel: +27(0)21 863 0033                           *
 *                           E-mail: info@azoteq.com                          *
 *                                                                            *
 * ========================================================================== *
 * @file        iqs9150-example-code.ino                                      *
 * @brief       IQS9150 Trackpad EV-KIT Example code                          *
 *              (PCB Version - AZP1364A1)                                     *
 *              This example demonstrates how to write the desired settings   *
 *              to the IQS9150 in order to use the IQS9150 Trackpad EV-KIT.   *
 *                                                                            *
 *              All data is displayed over serial communication with the      *
 *              following outputs:                                            *
 *                  - Gestures output(if enabled)                             *
 *                  - Finger 1: X and Y Coordinates                           *
 *                  - Finger 2: X and Y Coordinates                           *
 *                  - Finger 3: X and Y Coordinates                           *
 *                  - Power Mode Feedback                                     *
 *                  - Forced Communication                                    *
 *                                                                            *
 * @author      Azoteq PTY Ltd                                                *
 * @version     v1.0                                                          *
 * @date        2024-06-11                                                    *
 *****************************************************************************/

#include <Arduino.h>
#include <HID-Project.h>
#include <HID-Settings.h>
#include "src\IQS9150\IQS9150.h"

/*** Defines ***/
#define DEMO_IQS9150_ADDR                       0x56
#define DEMO_IQS9150_POWER_PIN                  4
#define DEMO_IQS9150_RDY_PIN                    7

/*** Instances ***/
IQS9150 iqs9150;

/*** Global Variables ***/
iqs9150_power_mode_e running_power_mode = IQS9150_ACTIVE_MODE;
bool show_data = false;
bool heading = true;
uint16_t running_x_output = 65535;
uint16_t running_y_output = 65535;
iqs9150_gestures_e running_gestures = IQS9150_GESTURE_NONE;


void setup()
{
  /* Start Serial Communication */
  Serial.begin(115200);
  while(!Serial);
  Serial.println("Start Serial communication");
  delay(200);

  /* Power On IQS9150 */
  pinMode(DEMO_IQS9150_POWER_PIN, OUTPUT);
  delay(200);
  digitalWrite(DEMO_IQS9150_POWER_PIN, LOW);
  delay(200);
  digitalWrite(DEMO_IQS9150_POWER_PIN, HIGH);
  delay(200);

  /* Initialize the IQS9150 with input parameters device address and RDY pin */
  iqs9150.begin(DEMO_IQS9150_ADDR, DEMO_IQS9150_RDY_PIN);
  Serial.println("IQS9150 Ready");
  delay(200);
}

void loop()
{
  iqs9150.run(); // Runs the IQS9150 program loop

  /* Function to initialize a force communication window. */
  force_comms_and_reset();

  /* If data was updated, display data read from IQS9150 */
  if (iqs9150.new_data_available)
  {
    /* Display the heading when a finger is lifted from the trackpad */
    printHeading();
   /* Print the following if the new data did not come from a force comms command */
    if (!show_data)
    {
      if(printData())
      {
        // printPowerMode();
        printGesture();
        printCoordinates();
        heading = true;
      }
    }
    /* Set this flag to false to indicate that the new data was already displayed/used */
    iqs9150.new_data_available = false;
    /* Check if a force comms command was sent and if we should display the data read */
    show_iqs9150_data();
  }
}

/* Force the IQS9150 to open a RDY window and read the current state of the
 * device or request a software reset */
void force_comms_and_reset(void)
{
  char message = read_message();

  /* If an 'f' was received over serial, open a forced communication window and
  print the new data received */
  if(message == 'f')
  {
    iqs9150.force_I2C_communication(); // prompt the IQS9150
    show_data = true;
  }

  /* If an 'r' was received over serial, request a software (SW) reset */
  if(message == 'r')
  {
    Serial.println("Software Reset Requested!");
    iqs9150.iqs9150_state.state = IQS9150_STATE_SW_RESET;
  }

  if(message == 'h')
  {
    printHeading();
  }
}

/* Read message sent over serial communication */
char read_message(void)
{
  while (Serial.available())
  {
    if(Serial.available() > 0)
    {
      return  Serial.read();
    }
  }

  return '\n';
}


/* Check Power mode and print out the current power mode */
void printPowerMode(void)
{
  iqs9150_power_mode_e buffer = iqs9150.getPowerMode();
  if(running_power_mode != buffer)
  {
    switch (buffer)
    {
      case IQS9150_ACTIVE_MODE:
        Serial.print("Active Mode");
        break;

      case IQS9150_IDLE_TOUCH_MODE:
        Serial.print("Idle-Touch Mode");
        break;

      case IQS9150_IDLE_MODE:
        Serial.print("Idle Mode");
        break;

      case IQS9150_LP1_MODE:
        Serial.print("LP 1 Mode");
        break;

      case IQS9150_LP2_MODE:
        Serial.print("LP 2 Mode");
        break;

      default:
        Serial.print("Unknown Mode");
        break;
      }

      /* Update the running power mode value with the buffer value */
      running_power_mode = buffer;
  }
  else
  {
    Serial.print("\t| ");
  }
}

/* Function to print heading for the Serial data display demo purposes */
void printHeading(void)
{
  /* Check if it is necessary to display the heading */
  if(heading && iqs9150.getNumFingers() == 0)
  {
    heading = false;
    Serial.println("\n\nGesture\t\t\t\t| 1 X\t| 1 Y\t| 2 X\t| 2 Y\t| 3 X\t| 3 Y\t| 4 X\t| 4 Y\t| 5 X\t| 5 Y\t| 6 X\t| 6 Y\t| 7 X\t| 7 Y\t|\n");
    // Serial.println("____________________________________________________________________________________________________________________");
  }
}

/* Function to determine if it is necessary to print all the relevant data
in the serial terminal */
bool printData(void)
{
/* See if it is necessary to display the button state, power mode or gestures */
  if(
      (
        (iqs9150.getPowerMode()       != running_power_mode)
      ||(iqs9150.getTouchpadEvent()   != running_gestures)
      ||(iqs9150.getAbsXCoordinate(1) != running_x_output)
      ||(iqs9150.getAbsYCoordinate(1) != running_y_output)
      )
    )
  {
    return true; // Let the main loop know to display the latest values
  }
  else
  {
    return false; // Let the main loop know to not display the latest values
  }

  return false;
}

/* Function to print the X and Y coordinates of finger 1 */
void printCoordinates(void)
{
  uint8_t numFingers = iqs9150.getNumFingers();

  for(int i = 1; i < numFingers+1; i++)
  {
    uint16_t buffer = iqs9150.getAbsXCoordinate(i);

    Serial.print(buffer);       // Print X coordinates
    running_x_output = buffer;
    Serial.print("\t|");       // Spacing to match headings in Print

    buffer = iqs9150.getAbsYCoordinate(i);

    Serial.print(buffer);       // Print Y coordinates
    running_y_output = buffer;
    Serial.print("\t|");       // Spacing to match headings in Print
  }
  Serial.println("");
}


/* Check if one of the gesture flags is set and display which one */
void printGesture(void)
{
  iqs9150_gestures_e buffer = iqs9150.getTouchpadEvent();

  switch (buffer)
  {
    case IQS9150_SINGLE_GESTURE_SINGLE_TAP:
      Serial.print("Single Tap\t\t\t|");
    break;

    case IQS9150_SINGLE_GESTURE_DOUBLE_TAP:
      Serial.print("Double Tap\t\t\t|");
    break;

    case IQS9150_SINGLE_GESTURE_TRIPLE_TAP:
      Serial.print("Triple Tap\t\t\t|");
    break;

    case IQS9150_SINGLE_GESTURE_PRESS_HOLD:
      Serial.print("Press and Hold\t\t\t|");
    break;

    case IQS9150_SINGLE_GESTURE_PALM_GESTURE:
      Serial.print("Palm Gesture\t\t\t|");
    break;

    case IQS9150_SINGLE_GESTURE_SWIPE_X_POSITIVE:
      Serial.print("Swipe X +\t\t\t|");
    break;

    case IQS9150_SINGLE_GESTURE_SWIPE_X_NEGATIVE:
      Serial.print("Swipe X -\t\t\t|");
    break;

    case IQS9150_SINGLE_GESTURE_SWIPE_Y_POSITIVE:
      Serial.print("Swipe Y +\t\t\t|");
    break;

    case IQS9150_SINGLE_GESTURE_SWIPE_Y_NEGATIVE:
      Serial.print("Swipe Y -\t\t\t|");
    break;

    case IQS9150_SINGLE_GESTURE_SWIPE_HOLD_X_POSITIVE:
      Serial.print("Swipe and Hold X +\t\t|");
    break;

    case IQS9150_SINGLE_GESTURE_SWIPE_HOLD_X_NEGATIVE:
      Serial.print("Swipe and Hold X -\t\t|");
    break;

    case IQS9150_SINGLE_GESTURE_SWIPE_HOLD_Y_POSITIVE:
      Serial.print("Swipe and Hold Y +\t\t|");
    break;

    case IQS9150_SINGLE_GESTURE_SWIPE_HOLD_Y_NEGATIVE:
      Serial.print("Swipe and Hold Y -\t\t|");
    break;

    case IQS9150_DOUBLE_GESTURE_SINGLE_TAP:
      Serial.print("Two Finger Single Tap\t\t|");
    break;

    case IQS9150_DOUBLE_GESTURE_DOUBLE_TAP:
      Serial.print("Two Finger Double Tap\t\t|");
    break;

    case IQS9150_DOUBLE_GESTURE_TRIPLE_TAP:
      Serial.print("Two Finger Triple Tap\t\t|");
    break;

    case IQS9150_DOUBLE_GESTURE_PRESS_HOLD:
      Serial.print("Two Finger Press and Hold\t|");
    break;

    case IQS9150_DOUBLE_GESTURE_ZOOM_IN:
      Serial.print("Zoom In\t\t\t|");
    break;

    case IQS9150_DOUBLE_GESTURE_ZOOM_OUT:
      Serial.print("Zoom Out\t\t\t|");
    break;

    case IQS9150_DOUBLE_GESTURE_VERTICAL_SCROLL:
      Serial.print("Vertical Scroll\t\t\t|");
    break;

    case IQS9150_DOUBLE_GESTURE_HORIZONTAL_SCROLL:
      Serial.print("Horizontal Scroll\t\t|");
    break;

    default:
      Serial.print("-\t\t\t\t|");
    break;
  }



  /* Update the running gesture value with the value from the buffer */
  running_gestures = buffer;
}

/* Shows the current IQS9150 data when a force comms command was sent */
void show_iqs9150_data()
{
  if (show_data)
  {
    Serial.println("********************* IQS9150 DATA - FORCED COMMS ***********************");
    Serial.println("Gesture:\t\tFinger 1 X:\tFinger 1 Y:\tPower Mode:");
    printGesture();
    printCoordinates();
    printPowerMode();
    show_data = false;
  }
}
