/******************************************************************************
 *                                                                            *
 *                                Copyright by                                *
 *                                                                            *
 *                              Azoteq (Pty) Ltd                              *
 *                          Republic of South Africa                          *
 *                                                                            *
 *                           Tel: +27(0)21 863 0033                           *
 *                           E-mail: info@azoteq.com                          *
 *                                                                            *
 * ========================================================================== *
 * @file        IQS9150.cpp                                                   *
 * @brief       This file contains the constructors and methods which allow   *
 *              ease of use of an IQS9150 Integrated Circuit (IC). The        *
 *              IQS9150 is a high resolution trackpad device capable of       *
 *              supporting up to 26x19 channels. This class                   *
 *              provides an easy means of initializing and interacting with   *
 *              the IQS9150 device from an Arduino-based device.              *
 * @author      CC van Eeden - Azoteq PTY Ltd                                 *
 * @version     v0.0.0                                                        *
 * @date        2024                                                          *
 * ========================================================================== *
 * @attention  Makes use of the following standard Arduino libraries:         *
 * - Arduino.h -> Included in IQS9150.h, comes standard with Arduino          *
 * - azq_i2c.h -> Included in IQS9150.h, Azoteq wrapper for Arduino's 'Wire'  *
 *****************************************************************************/

/* Include Files */
#include "IQS9150.h"
#include "IQS9150_AZP1364A1_init.h" //Replace this file if you are using a different hardware setup.


bool iqs9150_deviceRDY = false;
uint8_t iqs9150_rdy_pin;

void iqs9150_ready_interrupt(void);

/*****************************************************************************/
/*                             CONSTRUCTORS                                  */
/*****************************************************************************/
IQS9150::IQS9150(){
}

/*****************************************************************************/
/*                            PUBLIC METHODS                                 */
/*****************************************************************************/

/**
  * @name   begin
  * @brief  A method to initialize the IQS9150 device with the I2C Bus,
  *         device address and ready pin specified by the user.
  * @param  deviceAddressIn ->  The address of the IQS9150 device.
  * @param  rdy_pin      ->  The Arduino pin which is connected to the ready
  *                           pin of the IQS9150 device.
  * @param  clk_speed     ->  I2C Clock speed. Will force I2C bus to run at
  *                           defined speed. Can break other devices.
  * @retval None.
  * @note   - Receiving a true return value does not mean that initialization
  *           was successful.
  *         - Receiving a true return value only means that the IQS device
  *           responded to the request for communication.
  *         - Receiving a false return value means that initialization did not
  *           take place at all.
  *         - If communication is successfully established then it is unlikely
  *           that initialization will fail.
*/
void IQS9150::begin(uint8_t deviceAddressIn, uint8_t rdy_pin)
{
  // Initialize I2C communication here, since this library can't function without it.
  Wire.begin();
  Wire.setClock(400000);

  /* Initialize I2C communication here, since this library can't function
  without it. */
  _deviceAddress  = deviceAddressIn;
  iqs9150_rdy_pin       = rdy_pin;
  attachInterrupt(digitalPinToInterrupt(iqs9150_rdy_pin), iqs9150_ready_interrupt, CHANGE);

  /* Initialize "running" and "init" state machine variables. */
  iqs9150_state.state = IQS9150_STATE_START;
  iqs9150_state.init_state = IQS9150_INIT_VERIFY_PRODUCT;
}

/**
  * @name   init
  * @brief  A method that runs through a normal start-up routine to set up the
  *         IQS9150 with the desired settings from the IQS9150_init.h file.
  * @retval Returns true if the full start-up routine has been completed,
  *         returns false if not.
  * @note   - No false return will be given, the program will thus be stuck
  *           when one of the cases is not able to finish.
  *         - See serial communication to find the ERROR case
  */
bool IQS9150::init(void)
{
  uint16_t prod_num;
  uint8_t ver_maj, ver_min;

  switch (iqs9150_state.init_state)
  {
    /* Verifies product number to determine if the correct device is connected
    for this example */
    case IQS9150_INIT_VERIFY_PRODUCT:
      if(iqs9150_deviceRDY)
      {
        Serial.println("IQS9150_INIT_VERIFY_PRODUCT");
        prod_num = getProductNum(STOP);
        ver_maj = getmajorVersion(STOP);
        ver_min = getminorVersion(STOP);
        Serial.print("\t\tProduct number is: ");
        Serial.print(prod_num);
        Serial.print(" v");
        Serial.print(ver_maj);
        Serial.print(".");
        Serial.println(ver_min);
        if(prod_num == IQS9150_PRODUCT_NUM)
        {
          iqs9150_state.init_state = IQS9150_INIT_READ_RESET;
        }
        else
        {
          Serial.println("\t\tDevice is not a IQS9150!");
          iqs9150_state.init_state = IQS9150_INIT_NONE;
        }
      }
    break;

    /* Verify if a reset has occurred */
    case IQS9150_INIT_READ_RESET:
      if(iqs9150_deviceRDY)
      {
        Serial.println("IQS9150_INIT_READ_RESET");
        updateInfoFlags(STOP);
        if (checkReset())
        {
          Serial.println("\t\tReset event occurred.");
          iqs9150_state.init_state = IQS9150_INIT_ACK_RESET;
        }
        else
        {
          Serial.println("\t\t No Reset Event Detected - Request SW Reset");
          iqs9150_state.init_state = IQS9150_INIT_CHIP_RESET;
        }
      }
    break;

    /* Perform SW Reset */
    case IQS9150_INIT_CHIP_RESET:
      if(iqs9150_deviceRDY)
      {
        Serial.println("IQS9150_INIT_CHIP_RESET");

        //Perform SW Reset
        SW_Reset(STOP);
        Serial.println("\t\tSoftware Reset Bit Set.");
        delay(100);
        iqs9150_state.init_state = IQS9150_INIT_READ_RESET;
      }
    break;

    /* Acknowledge that the device went through a reset */
    case IQS9150_INIT_ACK_RESET:
      if(iqs9150_deviceRDY)
      {
        Serial.println("IQS9150_INIT_ACK_RESET");
        acknowledgeReset(STOP);
        delay(10);
        iqs9150_state.init_state = IQS9150_INIT_UPDATE_SETTINGS;
      }
      break;

    /* Write desired settings to the IQS9150 */
    case IQS9150_INIT_UPDATE_SETTINGS:
      if(iqs9150_deviceRDY)
      {
        Serial.println("IQS9150_INIT_UPDATE_SETTINGS");
        updateSettings(STOP);
        iqs9150_state.init_state = IQS9150_INIT_ATI;
      }
    break;

    /* Run the ATI algorithm to recalibrate the device with newly added settings */
    case IQS9150_INIT_ATI:
      if(iqs9150_deviceRDY)
      {
        Serial.println("IQS9150_INIT_ATI");
        ReATI();
        iqs9150_state.init_state = IQS9150_INIT_WAIT_FOR_ATI;
        Serial.println("IQS9150_INIT_WAIT_FOR_ATI");
      }
    break;

    /* Read the ATI Active bit to see if the rest of the program can continue */
    case IQS9150_INIT_WAIT_FOR_ATI:
      if(iqs9150_deviceRDY)
      {
        delay(10);
        if(readATIoccurred())
        {
          iqs9150_state.init_state = IQS9150_INIT_RESEED;
        }
      }
    break;

    /*Reseed */
    case IQS9150_INIT_RESEED:
      if(iqs9150_deviceRDY)
      {
        ReSeed();
        Serial.println("IQS9150_INIT_RESEED");
          iqs9150_state.init_state = IQS9150_INIT_ACTIVATE_EVENT_MODE;
      }
    break;

    case IQS9150_INIT_ACTIVATE_EVENT_MODE:
      if(iqs9150_deviceRDY)
      {
        Serial.println("IQS9150_INIT_ACTIVATE_EVENT_MODE");
        setEventMode();
        Serial.println("\t\tDONE");
        iqs9150_state.init_state = IQS9150_INIT_DONE;
      }
    break;


    /* Read the latest data from the iqs9150 */
    case IQS9150_INIT_READ_DATA:
      if(iqs9150_deviceRDY)
      {
        Serial.println("IQS9150_INIT_READ_DATA");
        queueValueUpdates();
        delay(10);
        iqs9150_state.init_state = IQS9150_INIT_DONE;
      }
    break;

    /* If all operations have been completed correctly, the RDY pin can be set
     * up as an interrupt to indicate when new data is available */
    case IQS9150_INIT_DONE:
      Serial.println("IQS9150_INIT_DONE\n");
      Serial.print("\e[s");
      new_data_available = true;
      return true;
    break;

    default:
      break;
  }
  return false;
}

/**
  * @name   run
  * @brief  This method is called continuously during runtime, which serves as
  *         the main state machine.
  * @param  None.
  * @retval None.
  * @note   The state machine continuously checks for specific events and
  *         updates the state machine accordingly. A reset event will cause
  *         the state machine to re-initialize the device.
  *
  *         queueValueUpdates can be edited by the user if other data should be
  *         read every time a RDY window is received.
  */
void IQS9150::run(void)
{
  switch (iqs9150_state.state)
  {
    /* After a hardware reset, this is the starting position of the main
    state machine */
    case IQS9150_STATE_START:
      iqs9150_state.state = IQS9150_STATE_INIT;
    break;

    /* Perform the initialization routine on the IQS9150 */
    case IQS9150_STATE_INIT:
      if(init())
      {
        iqs9150_state.state = IQS9150_STATE_RUN;
      }
    break;

    /* Send an I2C software reset in the next RDY window */
    case IQS9150_STATE_SW_RESET:
      SW_Reset(STOP);
      iqs9150_state.state = IQS9150_STATE_RUN;
    break;

    /* Continuous reset monitoring state, ensure no reset event has occurred
    for data to be valid */
    case IQS9150_STATE_CHECK_RESET:
      if(checkReset())
      {
        Serial.println("Reset Occurred!\n");
        new_data_available = false;
        iqs9150_state.state = IQS9150_STATE_START;
        iqs9150_state.init_state = IQS9150_INIT_VERIFY_PRODUCT;
      }

      /* A reset did not occur, move to the run state and wait for a new ready
      window */
      else
      {
        new_data_available = true; /* No reset, thus data is valid */
        iqs9150_state.state = IQS9150_STATE_RUN;
      }
    break;

    /* If a RDY Window is open, read the latest values from the IQS9150 */
    case IQS9150_STATE_RUN:
      if(iqs9150_deviceRDY)
      {
        queueValueUpdates();
        iqs9150_deviceRDY = false;
        new_data_available = false;
        iqs9150_state.state = IQS9150_STATE_CHECK_RESET;
      }
    break;
  }
}

/**
 * @name   iqs9150_ready_interrupt
 * @brief  A method used as an interrupt function. Only activated when a
 *         High -> Low (Falling edge) and LOW -> High (Rising edge) interrupt
 *         is seen on the correct Arduino interrupt pin.
 * @param  None.
 * @retval None.
 * @note   Keep this function as simple as possible to prevent stuck states
 *         and slow operations.
 */
void iqs9150_ready_interrupt(void)
{
    if(digitalRead(iqs9150_rdy_pin))
  {
    iqs9150_deviceRDY = false;
  }
  else
  {
    iqs9150_deviceRDY = true;
  }
}

/**
  * @name   clearRDY
  * @brief  A method used to clear the ready interrupt bit.
  * @param  None.
  * @retval None.
  */
void IQS9150::clearRDY(void)
{
  iqs9150_deviceRDY = false;
}

/**
  * @name   getRDYStatus
  * @brief  A method used to retrieve the device RDY status.
  * @param  None.
  * @retval Returns the boolean IQS9150 RDY state.
  *         - True when RDY line is LOW
  *         - False when RDY line is HIGH
  */
bool IQS9150::getRDYStatus(void)
{
  return iqs9150_deviceRDY;
}

/**
  * @name   requestData
  * @brief  Request the IQS9150 to perform its main loop and return new
  *         information from the device.
  * @param  None.
  * @retval None.
*/
void IQS9150::requestData(void)
{
  iqs9150_state.state = IQS9150_STATE_RUN;
}

/**
  * @name   queueValueUpdates
  * @brief  All I2C read operations in the queueValueUpdates method will be
  *         performed each time the IQS9150 opens a RDY window.
  * @param  None.
  * @retval None.
  * @note   Any Address in the IQS9150 memory map can be read from here.
  */
void IQS9150::queueValueUpdates(void)
{
  uint8_t transferBytes[30];	// The array which will hold the bytes to be transferred.

  readRandomBytes16(_deviceAddress, IQS9150_MM_SINGLE_FINGER_GESTURES, 8, transferBytes, RESTART);

  IQSMemoryMap.SINGLE_FINGER_GESTURES[0] = transferBytes[0];
  IQSMemoryMap.SINGLE_FINGER_GESTURES[1] = transferBytes[1];
  IQSMemoryMap.DOUBLE_FINGER_GESTURES[0] = transferBytes[2];
  IQSMemoryMap.DOUBLE_FINGER_GESTURES[1] = transferBytes[3];

	/* Read the info flags. 2 System flags bytes for activation and filter halt */
	// readOnly(_deviceAddress, 2+(2*bytes_per_field), transferBytes, STOP);

	/* Assign the System Status */
  IQSMemoryMap.INFO_FLAGS[0] =      transferBytes[4];
  IQSMemoryMap.INFO_FLAGS[1] =      transferBytes[5];
  IQSMemoryMap.TRACKPAD_FLAGS[0] =  transferBytes[6];
  IQSMemoryMap.TRACKPAD_FLAGS[1] =  transferBytes[7];

  // Determine how many fingers' data needs to be requested.
  uint8_t numTouches = IQSMemoryMap.TRACKPAD_FLAGS[0] & 0x0F;
  // Need to request at least one finger's data, if one or zero fingers are detected.
  if(numTouches <= 1)
  {
    readRandomBytes16(_deviceAddress, IQS9150_MM_FINGER_1_X, 8, transferBytes, STOP);
    IQSMemoryMap.FINGER_1_X[0] =              transferBytes[0];
    IQSMemoryMap.FINGER_1_X[1] =              transferBytes[1];
    IQSMemoryMap.FINGER_1_Y[0] =              transferBytes[2];
    IQSMemoryMap.FINGER_1_Y[1] =              transferBytes[3];
    IQSMemoryMap.FINGER_1_TOUCH_STRENGTH[0] = transferBytes[4];
    IQSMemoryMap.FINGER_1_TOUCH_STRENGTH[1] = transferBytes[5];
    IQSMemoryMap.FINGER_1_AREA[0] =           transferBytes[6];
    IQSMemoryMap.FINGER_1_AREA[1] =           transferBytes[7];
  }
  else
  {
    /*  If less than four fingers are detected on trackpad it can be requested
        in one read block, otherwise it needs to be requested in two blocks*/
    if(numTouches < 4)
    {
      readRandomBytes16(_deviceAddress, IQS9150_MM_FINGER_1_X, 8*numTouches, transferBytes, STOP);
      IQSMemoryMap.FINGER_1_X[0] =              transferBytes[0];
      IQSMemoryMap.FINGER_1_X[1] =              transferBytes[1];
      IQSMemoryMap.FINGER_1_Y[0] =              transferBytes[2];
      IQSMemoryMap.FINGER_1_Y[1] =              transferBytes[3];
      IQSMemoryMap.FINGER_1_TOUCH_STRENGTH[0] = transferBytes[4];
      IQSMemoryMap.FINGER_1_TOUCH_STRENGTH[1] = transferBytes[5];
      IQSMemoryMap.FINGER_1_AREA[0] =           transferBytes[6];
      IQSMemoryMap.FINGER_1_AREA[1] =           transferBytes[7];
      IQSMemoryMap.FINGER_2_X[0] =              transferBytes[8];
      IQSMemoryMap.FINGER_2_X[1] =              transferBytes[9];
      IQSMemoryMap.FINGER_2_Y[0] =              transferBytes[10];
      IQSMemoryMap.FINGER_2_Y[1] =              transferBytes[11];
      IQSMemoryMap.FINGER_2_TOUCH_STRENGTH[0] = transferBytes[12];
      IQSMemoryMap.FINGER_2_TOUCH_STRENGTH[1] = transferBytes[13];
      IQSMemoryMap.FINGER_2_AREA[0] =           transferBytes[14];
      IQSMemoryMap.FINGER_2_AREA[1] =           transferBytes[15];
      IQSMemoryMap.FINGER_3_X[0] =              transferBytes[16];
      IQSMemoryMap.FINGER_3_X[1] =              transferBytes[17];
      IQSMemoryMap.FINGER_3_Y[0] =              transferBytes[18];
      IQSMemoryMap.FINGER_3_Y[1] =              transferBytes[19];
      IQSMemoryMap.FINGER_3_TOUCH_STRENGTH[0] = transferBytes[20];
      IQSMemoryMap.FINGER_3_TOUCH_STRENGTH[1] = transferBytes[21];
      IQSMemoryMap.FINGER_3_AREA[0] =           transferBytes[22];
      IQSMemoryMap.FINGER_3_AREA[1] =           transferBytes[23];
    }
    else
    {
      readRandomBytes16(_deviceAddress, IQS9150_MM_FINGER_1_X, 30, transferBytes, RESTART);
      IQSMemoryMap.FINGER_1_X[0] =              transferBytes[0];
      IQSMemoryMap.FINGER_1_X[1] =              transferBytes[1];
      IQSMemoryMap.FINGER_1_Y[0] =              transferBytes[2];
      IQSMemoryMap.FINGER_1_Y[1] =              transferBytes[3];
      IQSMemoryMap.FINGER_1_TOUCH_STRENGTH[0] = transferBytes[4];
      IQSMemoryMap.FINGER_1_TOUCH_STRENGTH[1] = transferBytes[5];
      IQSMemoryMap.FINGER_1_AREA[0] =           transferBytes[6];
      IQSMemoryMap.FINGER_1_AREA[1] =           transferBytes[7];
      IQSMemoryMap.FINGER_2_X[0] =              transferBytes[8];
      IQSMemoryMap.FINGER_2_X[1] =              transferBytes[9];
      IQSMemoryMap.FINGER_2_Y[0] =              transferBytes[10];
      IQSMemoryMap.FINGER_2_Y[1] =              transferBytes[11];
      IQSMemoryMap.FINGER_2_TOUCH_STRENGTH[0] = transferBytes[12];
      IQSMemoryMap.FINGER_2_TOUCH_STRENGTH[1] = transferBytes[13];
      IQSMemoryMap.FINGER_2_AREA[0] =           transferBytes[14];
      IQSMemoryMap.FINGER_2_AREA[1] =           transferBytes[15];
      IQSMemoryMap.FINGER_3_X[0] =              transferBytes[16];
      IQSMemoryMap.FINGER_3_X[1] =              transferBytes[17];
      IQSMemoryMap.FINGER_3_Y[0] =              transferBytes[18];
      IQSMemoryMap.FINGER_3_Y[1] =              transferBytes[19];
      IQSMemoryMap.FINGER_3_TOUCH_STRENGTH[0] = transferBytes[20];
      IQSMemoryMap.FINGER_3_TOUCH_STRENGTH[1] = transferBytes[21];
      IQSMemoryMap.FINGER_3_AREA[0] =           transferBytes[22];
      IQSMemoryMap.FINGER_3_AREA[1] =           transferBytes[23];
      IQSMemoryMap.FINGER_4_X[0] =              transferBytes[24];
      IQSMemoryMap.FINGER_4_X[1] =              transferBytes[25];
      IQSMemoryMap.FINGER_4_Y[0] =              transferBytes[26];
      IQSMemoryMap.FINGER_4_Y[1] =              transferBytes[27];
      IQSMemoryMap.FINGER_4_TOUCH_STRENGTH[0] = transferBytes[28];
      IQSMemoryMap.FINGER_4_TOUCH_STRENGTH[1] = transferBytes[29];

      readRandomBytes16(_deviceAddress, IQS9150_MM_FINGER_4_AREA, 8*numTouches - 30, transferBytes, STOP);
      IQSMemoryMap.FINGER_4_AREA[0] =           transferBytes[0];
      IQSMemoryMap.FINGER_4_AREA[1] =           transferBytes[1];
      IQSMemoryMap.FINGER_5_X[0] =              transferBytes[2];
      IQSMemoryMap.FINGER_5_X[1] =              transferBytes[3];
      IQSMemoryMap.FINGER_5_Y[0] =              transferBytes[4];
      IQSMemoryMap.FINGER_5_Y[1] =              transferBytes[5];
      IQSMemoryMap.FINGER_5_TOUCH_STRENGTH[0] = transferBytes[6];
      IQSMemoryMap.FINGER_5_TOUCH_STRENGTH[1] = transferBytes[7];
      IQSMemoryMap.FINGER_5_AREA[0] =           transferBytes[8];
      IQSMemoryMap.FINGER_5_AREA[1] =           transferBytes[9];
      IQSMemoryMap.FINGER_6_X[0] =              transferBytes[10];
      IQSMemoryMap.FINGER_6_X[1] =              transferBytes[11];
      IQSMemoryMap.FINGER_6_Y[0] =              transferBytes[12];
      IQSMemoryMap.FINGER_6_Y[1] =              transferBytes[13];
      IQSMemoryMap.FINGER_6_TOUCH_STRENGTH[0] = transferBytes[14];
      IQSMemoryMap.FINGER_6_TOUCH_STRENGTH[1] = transferBytes[15];
      IQSMemoryMap.FINGER_6_AREA[0] =           transferBytes[16];
      IQSMemoryMap.FINGER_6_AREA[1] =           transferBytes[17];
      IQSMemoryMap.FINGER_7_X[0] =              transferBytes[18];
      IQSMemoryMap.FINGER_7_X[1] =              transferBytes[19];
      IQSMemoryMap.FINGER_7_Y[0] =              transferBytes[20];
      IQSMemoryMap.FINGER_7_Y[1] =              transferBytes[21];
      IQSMemoryMap.FINGER_7_TOUCH_STRENGTH[0] = transferBytes[22];
      IQSMemoryMap.FINGER_7_TOUCH_STRENGTH[1] = transferBytes[23];
      IQSMemoryMap.FINGER_7_AREA[0] =           transferBytes[24];
      IQSMemoryMap.FINGER_7_AREA[1] =           transferBytes[25];
    }
  }
	/* READ WRITE */
}

/**
  * @name	  readATIactive
  * @brief  A method that checks if the ATI routine has been completed
  * @param  tpOrAlp ->  Specifies what ATI bit to set. If true Trackpad ReATI bit
  *                     will be set, if false ALP ReATI bit will be set.
  *                     Defaults to Trackpad if not supplied.
  * @retval Returns true if the Re-ATI Occurred bit is set, false if the
  *         Re-ATI Occurred bit is set.
  * @note   If the ATI routine is active the channel states (NONE, PROX, TOUCH)
  *         might exhibit unwanted behaviour. Thus it is advised to wait for
  *         the routine to complete before continuing.
  */
bool IQS9150::readATIoccurred(bool tpOrAlp = TP)
{
  /* Read the Info flags */
  updateInfoFlags(STOP);

  /* Return the ATI Occurred status */
  if(tpOrAlp)
  {
    return getBit(IQSMemoryMap.INFO_FLAGS[0], IQS9150_TP_RE_ATI_OCCURRED_BIT);
  }
  else
  {
    return getBit(IQSMemoryMap.INFO_FLAGS[0], IQS9150_ALP_RE_ATI_OCCURRED_BIT);
  }

}

/**
  * @name	  checkReset
  * @brief  A method that checks if the device has reset and returns the reset
  *         status.
  * @param  None.
  * @retval Returns true if a reset has occurred, false if no reset has occurred.
  * @note   If a reset has occurred the device settings should be reloaded using
  *         the begin function. After new device settings have been reloaded
  *         the acknowledge reset function can be used to clear the reset flag.
  */
bool IQS9150::checkReset(void)
{
	/* Perform a bitwise AND operation with the SHOW_RESET_BIT to return the
  reset status */
  return getBit(IQSMemoryMap.INFO_FLAGS[0], IQS9150_SHOW_RESET_BIT);
}

/**
  * @name	  checkProductNum
  * @brief  A method that checks the device product number and compares the
  *         result to the defined value to return a boolean result.
  * @param  stopOrRestart ->  Specifies whether the communications window must
  *                           be kept open or must be closed after this action.
  *                           Use the STOP and RESTART definitions.
  * @retval Returns the product number as a unit16_t value.
  * @note   If the product is not correctly identified an appropriate messages
  *         should be displayed.
  */
uint16_t IQS9150::getProductNum(bool stopOrRestart)
{
	uint8_t transferBytes[2];	      // A temporary array to hold the byte to be transferred.
  uint8_t prodNumLow = 0;         // Temporary storage for the Counts low byte.
  uint8_t prodNumHigh = 0;        // Temporary storage for the Counts high byte.
  uint16_t prodNumReturn = 0;     // The 16bit return value.

	/* Read the Device info from the IQS9150. */
	readRandomBytes16(_deviceAddress, IQS9150_MM_PROD_NUM, 2, transferBytes, stopOrRestart);

  /* Construct the 16bit return value. */
  prodNumLow = transferBytes[0];
  prodNumHigh = transferBytes[1];
  prodNumReturn = (uint16_t)(prodNumLow);
  prodNumReturn |= (uint16_t)(prodNumHigh<<8);
  /* Return the Counts value. */
  return prodNumReturn;
}

/**
  * @name	getmajorVersion
  * @brief  A method that checks the device firmware version's major value.
  * @param  stopOrRestart -> Specifies whether the communications window must
  *                           be kept open or must be closed after this action.
  *                          Use the STOP and RESTART definitions.
  * @retval Returns major version number as a uint8_t value.
  */
uint8_t IQS9150::getmajorVersion(bool stopOrRestart)
{
	uint8_t transferBytes[2];	// A temporary array to hold the byte to be transferred.
  uint8_t ver_maj = 0;      // Temporary storage for the firmware version major number.

	/* Read the Device info from the IQS9150. */
	readRandomBytes16(_deviceAddress, IQS9150_MM_MAJOR_VERSION_NUM, 2, transferBytes, stopOrRestart);

  /* Get major value from correct byte */
  ver_maj = transferBytes[0];
  /* Return the major firmware version number value. */
  return ver_maj;
}

/**
  * @name	getminorVersion
  * @brief  A method that checks the device firmware version's minor value.
  * @param  stopOrRestart ->  Specifies whether the communications window must
  *                           be kept open or must be closed after this action.
  *                           Use the STOP and RESTART definitions.
  * @retval Returns minor version number as a unit8_t value.
  */
uint8_t IQS9150::getminorVersion(bool stopOrRestart)
{
	uint8_t transferBytes[2];	// A temporary array to hold the byte to be transferred.
  uint8_t ver_min = 0;      // Temporary storage for the firmware version minor number.

	/* Read the Device info from the IQS9150. */
	readRandomBytes16(_deviceAddress, IQS9150_MM_MINOR_VERSION_NUM, 2, transferBytes, stopOrRestart);
  /* get major value from correct byte */
  ver_min = transferBytes[0];
  /* Return the minor firmware version number value. */
  return ver_min;
}

/**
  * @name	acknowledgeReset
  * @brief  A method that clears the Reset Event bit by writing it to a 0.
  * @param  stopOrRestart ->  Specifies whether the communications window must
  *                           be kept open or must be closed after this action.
  *                           Use the STOP and RESTART definitions.
  * @retval None.
  * @note   If a reset has occurred the device settings should be reloaded using
  *         the begin function. After new device settings have been reloaded
  *         this method should be used to clear the reset bit.
  */
void IQS9150::acknowledgeReset(bool stopOrRestart)
{
	uint8_t transferByte[2];	// A temporary array to hold the bytes to be transferred.
	/* Read the System Flags from the IQS9150, these must be read first in order
  not to change any settings. */
	readRandomBytes16(_deviceAddress, IQS9150_MM_SYSTEM_CONTROL, 2, transferByte, STOP);
	/* Set the the ACK_RESET_BIT in the SYSTEM_CONTROL register */
	transferByte[0] = setBit(transferByte[0], IQS9150_ACK_RESET_BIT);
	/* Write the new byte to the System Flags address. */
	writeRandomBytes16(_deviceAddress, IQS9150_MM_SYSTEM_CONTROL, 2, transferByte, stopOrRestart);
}

/**
  * @name   ReATI
  * @brief  A method which sets the REDO_ATI_BIT in order to force the IQS9150
  *         device to run the Automatic Tuning Implementation (ATI) routine.
  * @retval None.
  * @note   I2C communications are disabled for the duration of the ATI process.
  */
void IQS9150::ReATI(bool tpOrAlp)
{
  uint8_t transferByte[2]; // Array to store the bytes transferred.

  readRandomBytes16(_deviceAddress, IQS9150_MM_SYSTEM_CONTROL, 2, transferByte, RESTART);


  if(tpOrAlp)
  {
    transferByte[0] = setBit(transferByte[0], IQS9150_TP_RE_ATI_BIT); /* Trackpad ATI */
  }
  else
  {
    transferByte[0] = setBit(transferByte[0], IQS9150_ALP_RE_ATI_BIT); /* ALP ATI */
  }
  writeRandomBytes16(_deviceAddress, IQS9150_MM_SYSTEM_CONTROL, 2, transferByte, STOP);
}

/**
  * @name   ReSeed
  * @brief  A method which sets the IQS9150_RESEED_BIT in order to force the IQS9150
  *         device to reseed the counts.
  * @retval None.
  * @note   I2C communications are disabled for the duration of the ATI process.
  */
void IQS9150::ReSeed(bool tpOrAlp)
{
  uint8_t transferByte[2]; // Array to store the bytes transferred.

  readRandomBytes16(_deviceAddress, IQS9150_MM_SYSTEM_CONTROL, 2, transferByte, RESTART);
  /* Set the the IQS9150_RESEED_BIT in the SYSTEM_CONTROL register */
  if(tpOrAlp)
  {
    transferByte[0] = setBit(transferByte[0], IQS9150_TP_RESEED_BIT);
  }
  else
  {
    transferByte[0] = setBit(transferByte[0], IQS9150_TP_RESEED_BIT);
  }
  /* Write the new byte to the required device. */
  writeRandomBytes16(_deviceAddress, IQS9150_MM_SYSTEM_CONTROL, 2, transferByte, STOP);
}

/**
  * @name   setEventMode
  * @brief  A method which sets the IQS9150_EVENT_MODE_BIT in order to force the IQS9150
  *         device into event mode.
  * @retval None.
  */
void IQS9150::setEventMode(void)
{
  uint8_t transferByte[2]; // Array to store the bytes transferred.

  readRandomBytes16(_deviceAddress, IQS9150_MM_CONFIG_SETTINGS, 2, transferByte, RESTART);
  /* Set the the IQS9150_EVENT_MODE_BIT in the CONFIG_SETTINGS register */
  transferByte[1] = setBit(transferByte[1], IQS9150_EVENT_MODE_BIT);
  /* Write the new byte to the required device. */
  writeRandomBytes16(_deviceAddress, IQS9150_MM_CONFIG_SETTINGS, 2, transferByte, STOP);
}

/**
  * @name   SW_Reset
  * @brief  A method that sets the SW RESET bit to force the IQS9150 device to
  *         do a software reset.
  * @param  stopOrRestart ->  Specifies whether the communications window must
  *                           be kept open or must be closed after this action.
  *                           Use the STOP and RESTART definitions.
  * @retval None.
  * @note   To perform SW Reset, bit IQS9150_SW_RESET_BIT in SYSTEM_CONTROL is set.
  */
void IQS9150::SW_Reset(bool stopOrRestart)
{
  uint8_t transferByte[2]; // Array to store the bytes transferred.

  readRandomBytes16(_deviceAddress, IQS9150_MM_SYSTEM_CONTROL, 2, transferByte, STOP);
  /* Set the the SW_RESET_BIT in the SYSTEM_CONTROL register */
  transferByte[1] = setBit(transferByte[0], IQS9150_SW_RESET_BIT);
  /* Write the new byte to the required device. */
  writeRandomBytes16(_deviceAddress, IQS9150_MM_SYSTEM_CONTROL, 2, transferByte, stopOrRestart);
}

/**
  * @name   updateInfoFlags
  * @brief  A method that reads the info flags from the IQS9150 and assigns
  *         it to the local INFO_FLAGS register.
  * @param  stopOrRestart ->  Specifies whether the communications window must
  *                           be kept open or must be closed after this action.
  *              			        Use the STOP and RESTART definitions.
  * @retval None.
  * @note   The local INFO_FLAGS register is altered with the new value of
  *         the info flags register, retrieved from the IQS9150.
  */
void IQS9150::updateInfoFlags(bool stopOrRestart)
{
	uint8_t transferBytes[2];	// The array which will hold the bytes to be transferred.

	/* Read the info flags. */
	readRandomBytes16(_deviceAddress, IQS9150_MM_INFO_FLAGS, 2, transferBytes, stopOrRestart);
	/* Assign the info flags to the local SYSTEM_STATUS register */
  IQSMemoryMap.INFO_FLAGS[0] =  transferBytes[0];
  IQSMemoryMap.INFO_FLAGS[1] =  transferBytes[1];
}

/**
  * @name   updateSettings
  * @brief  A method that writes in the settings to set up the device.
  * @param  stopOrRestart ->  Specifies whether the communications window must
  *                           be kept open or must be closed after this action.
  *              			        Use the STOP and RESTART definitions.
  * @retval None.
  * @note   The local IQSMemoryMap registers should be altered with new values
  *         before this function is called.
  */
void IQS9150::updateSettings(bool stopOrRestart)
{
  uint8_t transferBytes[30];

  /* Change the ALP ATI Compensation */
  /* Memory Map Position 0x115C - 0x1175 */
  transferBytes[0] = ALP_COMPENSATION_RX0_0;
  transferBytes[1] = ALP_COMPENSATION_RX0_1;
  transferBytes[2] = ALP_COMPENSATION_RX1_0;
  transferBytes[3] = ALP_COMPENSATION_RX1_1;
  transferBytes[4] = ALP_COMPENSATION_RX2_0;
  transferBytes[5] = ALP_COMPENSATION_RX2_1;
  transferBytes[6] = ALP_COMPENSATION_RX3_0;
  transferBytes[7] = ALP_COMPENSATION_RX3_1;
  transferBytes[8] = ALP_COMPENSATION_RX4_0;
  transferBytes[9] = ALP_COMPENSATION_RX4_1;
  transferBytes[10] = ALP_COMPENSATION_RX5_0;
  transferBytes[11] = ALP_COMPENSATION_RX5_1;
  transferBytes[12] = ALP_COMPENSATION_RX6_0;
  transferBytes[13] = ALP_COMPENSATION_RX6_1;
  transferBytes[14] = ALP_COMPENSATION_RX7_0;
  transferBytes[15] = ALP_COMPENSATION_RX7_1;
  transferBytes[16] = ALP_COMPENSATION_RX8_0;
  transferBytes[17] = ALP_COMPENSATION_RX8_1;
  transferBytes[18] = ALP_COMPENSATION_RX9_0;
  transferBytes[19] = ALP_COMPENSATION_RX9_1;
  transferBytes[20] = ALP_COMPENSATION_RX10_0;
  transferBytes[21] = ALP_COMPENSATION_RX10_1;
  transferBytes[22] = ALP_COMPENSATION_RX11_0;
  transferBytes[23] = ALP_COMPENSATION_RX11_1;
  transferBytes[24] = ALP_COMPENSATION_RX12_0;
  transferBytes[25] = ALP_COMPENSATION_RX12_1;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_ALP_COMPENSATION_RX0_0, 26, transferBytes, RESTART);

  /* Change the I2C Slave Address */
  /* Memory Map Position 0x1176 - 0x1177 */
  transferBytes[0] = I2C_KEY;
  transferBytes[1] = I2C_SLAVE_ADR;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_I2C_KEY, 2, transferBytes, RESTART);

  /* Change the Settings Version Numbers */
  /* Memory Map Position 0x1178 - 0x1179 */
  transferBytes[0] = MINOR_VERSION;
  transferBytes[1] = MAJOR_VERSION;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_MINOR_VERSION, 2, transferBytes, RESTART);

  /* Change the ATI Multipliers / Dividers */
  /* Memory Map Position 0x117A - 0x1195 */
  transferBytes[0] = TP_ATI_MULTDIV_L;
  transferBytes[1] = TP_ATI_MULTDIV_H;
  transferBytes[2] = ALP_ATI_COARSE_RX0_L;
  transferBytes[3] = ALP_ATI_COARSE_RX0_H;
  transferBytes[4] = ALP_ATI_COARSE_RX1_L;
  transferBytes[5] = ALP_ATI_COARSE_RX1_H;
  transferBytes[6] = ALP_ATI_COARSE_RX2_L;
  transferBytes[7] = ALP_ATI_COARSE_RX2_H;
  transferBytes[8] = ALP_ATI_COARSE_RX3_L;
  transferBytes[9] = ALP_ATI_COARSE_RX3_H;
  transferBytes[10] = ALP_ATI_COARSE_RX4_L;
  transferBytes[11] = ALP_ATI_COARSE_RX4_H;
  transferBytes[12] = ALP_ATI_COARSE_RX5_L;
  transferBytes[13] = ALP_ATI_COARSE_RX5_H;
  transferBytes[14] = ALP_ATI_COARSE_RX6_L;
  transferBytes[15] = ALP_ATI_COARSE_RX6_H;
  transferBytes[16] = ALP_ATI_COARSE_RX7_L;
  transferBytes[17] = ALP_ATI_COARSE_RX7_H;
  transferBytes[18] = ALP_ATI_COARSE_RX8_L;
  transferBytes[19] = ALP_ATI_COARSE_RX8_H;
  transferBytes[20] = ALP_ATI_COARSE_RX9_L;
  transferBytes[21] = ALP_ATI_COARSE_RX9_H;
  transferBytes[22] = ALP_ATI_COARSE_RX10_L;
  transferBytes[23] = ALP_ATI_COARSE_RX10_H;
  transferBytes[24] = ALP_ATI_COARSE_RX11_L;
  transferBytes[25] = ALP_ATI_COARSE_RX11_H;
  transferBytes[26] = ALP_ATI_COARSE_RX12_L;
  transferBytes[27] = ALP_ATI_COARSE_RX12_H;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_TP_ATI_MULTDIV_L, 28, transferBytes, RESTART);

  /* Change the ATI Settings */
  /* Memory Map Position 0x1196 - 0x11A1 */
  transferBytes[0] = TP_ATI_TARGET_0;
  transferBytes[1] = TP_ATI_TARGET_1;
  transferBytes[2] = ALP_ATI_TARGET_0;
  transferBytes[3] = ALP_ATI_TARGET_1;
  transferBytes[4] = ALP_BASE_TARGET_0;
  transferBytes[5] = ALP_BASE_TARGET_1;
  transferBytes[6] = TP_NEG_DELTA_REATI_0;
  transferBytes[7] = TP_NEG_DELTA_REATI_1;
  transferBytes[8] = TP_POS_DELTA_REATI_0;
  transferBytes[9] = TP_POS_DELTA_REATI_1;
  transferBytes[10] = TP_REF_DRIFT_LIMIT;
  transferBytes[11] = ALP_LTA_DRIFT_LIMIT;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_TP_ATI_TARGET_0, 12, transferBytes, RESTART);

  /* Change the Sampling Periods and Timing */
  /* Memory Map Position 0x11A2 - 0x11BB */
  transferBytes[0] = ACTIVE_MODE_SAMPLING_PERIOD_0;
  transferBytes[1] = ACTIVE_MODE_SAMPLING_PERIOD_1;
  transferBytes[2] = IDLE_TOUCH_MODE_SAMPLING_PERIOD_0;
  transferBytes[3] = IDLE_TOUCH_MODE_SAMPLING_PERIOD_1;
  transferBytes[4] = IDLE_MODE_SAMPLING_PERIOD_0;
  transferBytes[5] = IDLE_MODE_SAMPLING_PERIOD_1;
  transferBytes[6] = LP1_MODE_SAMPLING_PERIOD_0;
  transferBytes[7] = LP1_MODE_SAMPLING_PERIOD_1;
  transferBytes[8] = LP2_MODE_SAMPLING_PERIOD_0;
  transferBytes[9] = LP2_MODE_SAMPLING_PERIOD_1;
  transferBytes[10] = STATIONARY_TOUCH_TIMEOUT_0;
  transferBytes[11] = STATIONARY_TOUCH_TIMEOUT_1;
  transferBytes[12] = IDLE_TOUCH_MODE_TIMEOUT_0;
  transferBytes[13] = IDLE_TOUCH_MODE_TIMEOUT_1;
  transferBytes[14] = IDLE_MODE_TIMEOUT_0;
  transferBytes[15] = IDLE_MODE_TIMEOUT_1;
  transferBytes[16] = LP1_MODE_TIMEOUT_0;
  transferBytes[17] = LP1_MODE_TIMEOUT_1;
  transferBytes[18] = ACTIVE_MODE_TIMEOUT_0;
  transferBytes[19] = ACTIVE_MODE_TIMEOUT_1;
  transferBytes[20] = REATI_RETRY_TIME;
  transferBytes[21] = REF_UPDATE_TIME;
  transferBytes[22] = I2C_TIMEOUT_0;
  transferBytes[23] = I2C_TIMEOUT_1;
  transferBytes[24] = SNAP_TIMEOUT;
  transferBytes[25] = OPEN_TIMING;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_ACTIVE_MODE_SAMPLING_PERIOD_0, 26, transferBytes, RESTART);

  /* Change the System Settings */
  /* Memory Map Position 0x11BC - 0x11C1 */
  transferBytes[0] = SYSTEM_CONTROL_0;
  transferBytes[1] = SYSTEM_CONTROL_1;
  transferBytes[2] = CONFIG_SETTINGS_0;
  transferBytes[3] = CONFIG_SETTINGS_1;
  transferBytes[4] = OTHER_SETTINGS_0;
  transferBytes[5] = OTHER_SETTINGS_1;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_SYSTEM_CONTROL_0, 6, transferBytes, RESTART);

  /* Change the ALP Settings */
  /* Memory Map Position 0x11C2 - 0x11CB */
  transferBytes[0] = ALP_SETUP_0;
  transferBytes[1] = ALP_SETUP_1;
  transferBytes[2] = ALP_SETUP_2;
  transferBytes[3] = ALP_SETUP_3;
  transferBytes[4] = ALP_TX_ENABLE_0;
  transferBytes[5] = ALP_TX_ENABLE_1;
  transferBytes[6] = ALP_TX_ENABLE_2;
  transferBytes[7] = ALP_TX_ENABLE_3;
  transferBytes[8] = ALP_TX_ENABLE_4;
  transferBytes[9] = ALP_TX_ENABLE_5;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_ALP_SETUP_0, 10, transferBytes, RESTART);

  /* Change the Thresholds and Debounce Settings */
  /* Memory Map Position 0x11CC - 0x11D3 */
  transferBytes[0] = TRACKPAD_TOUCH_SET_THRESHOLD;
  transferBytes[1] = TRACKPAD_TOUCH_CLEAR_THRESHOLD;
  transferBytes[2] = ALP_THRESHOLD;
  transferBytes[3] = ALP_AUTOPROX_THRESHOLD;
  transferBytes[4] = ALP_SET_DEBOUNCE;
  transferBytes[5] = ALP_CLEAR_DEBOUNCE;
  transferBytes[6] = SNAP_SET_THRESHOLD;
  transferBytes[7] = SNAP_CLEAR_THRESHOLD;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_TRACKPAD_TOUCH_SET_THRESHOLD, 8, transferBytes, RESTART);

  /* Change the ALP Count and LTA Betas */
  /* Memory Map Position 0x11D4 - 0x11D7 */
  transferBytes[0] = ALP_COUNT_BETA_LP1;
  transferBytes[1] = ALP_LTA_BETA_LP1;
  transferBytes[2] = ALP_COUNT_BETA_LP2;
  transferBytes[3] = ALP_LTA_BETA_LP2;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_ALP_COUNT_BETA_LP1, 4, transferBytes, RESTART);

  /* Change the Hardware Settings */
  /* Memory Map Position 0x11D8 - 0x11E1 */
  transferBytes[0] = TP_FRAC;
  transferBytes[1] = TP_PERIOD1;
  transferBytes[2] = TP_PERIOD2;
  transferBytes[3] = ALP_FRAC;
  transferBytes[4] = ALP_PERIOD1;
  transferBytes[5] = ALP_PERIOD2;
  transferBytes[6] = TRACKPAD_HARDWARE_SETTINGS_0;
  transferBytes[7] = TRACKPAD_HARDWARE_SETTINGS_1;
  transferBytes[8] = ALP_HARDWARE_SETTINGS_0;
  transferBytes[9] = ALP_HARDWARE_SETTINGS_1;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_TP_FRAC, 10, transferBytes, RESTART);

  /* Change the Trackpad Settings */
  /* Memory Map Position 0x11E2 - 0x11F5 */
  transferBytes[0] = TRACKPAD_SETTINGS_0_0;
  transferBytes[1] = TRACKPAD_SETTINGS_0_1;
  transferBytes[2] = TRACKPAD_SETTINGS_1_0;
  transferBytes[3] = TRACKPAD_SETTINGS_1_1;
  transferBytes[4] = X_RESOLUTION_0;
  transferBytes[5] = X_RESOLUTION_1;
  transferBytes[6] = Y_RESOLUTION_0;
  transferBytes[7] = Y_RESOLUTION_1;
  transferBytes[8] = XY_DYNAMIC_FILTER_BOTTOM_SPEED_0;
  transferBytes[9] = XY_DYNAMIC_FILTER_BOTTOM_SPEED_1;
  transferBytes[10] = XY_DYNAMIC_FILTER_TOP_SPEED_0;
  transferBytes[11] = XY_DYNAMIC_FILTER_TOP_SPEED_1;
  transferBytes[12] = XY_DYNAMIC_FILTER_BOTTOM_BETA;
  transferBytes[13] = XY_DYNAMIC_FILTER_STATIC_FILTER_BETA;
  transferBytes[14] = STATIONARY_TOUCH_MOV_THRESHOLD;
  transferBytes[15] = FINGER_SPLIT_FACTOR;
  transferBytes[16] = X_TRIM_VALUE;
  transferBytes[17] = Y_TRIM_VALUE;
  transferBytes[18] = JITTER_FILTER_DELTA;
  transferBytes[19] = FINGER_CONFIDENCE_THRESHOLD;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_TRACKPAD_SETTINGS_0_0, 20, transferBytes, RESTART);

  /* Change the Gesture Settings */
  /* Memory Map Position 0x11F6 - 0x1217 */
  transferBytes[0] = GESTURE_ENABLE_0;
  transferBytes[1] = GESTURE_ENABLE_1;
  transferBytes[2] = GESTURE_ENABLE_2F_0;
  transferBytes[3] = GESTURE_ENABLE_2F_1;
  transferBytes[4] = TAP_TOUCH_TIME_0;
  transferBytes[5] = TAP_TOUCH_TIME_1;
  transferBytes[6] = TAP_WAIT_TIME_0;
  transferBytes[7] = TAP_WAIT_TIME_1;
  transferBytes[8] = TAP_DISTANCE_0;
  transferBytes[9] = TAP_DISTANCE_1;
  transferBytes[10] = HOLD_TIME_0;
  transferBytes[11] = HOLD_TIME_1;
  transferBytes[12] = SWIPE_TIME_0;
  transferBytes[13] = SWIPE_TIME_1;
  transferBytes[14] = SWIPE_X_DISTANCE_0;
  transferBytes[15] = SWIPE_X_DISTANCE_1;
  transferBytes[16] = SWIPE_Y_DISTANCE_0;
  transferBytes[17] = SWIPE_Y_DISTANCE_1;
  transferBytes[18] = SWIPE_X_CONS_DIST_0;
  transferBytes[19] = SWIPE_X_CONS_DIST_1;
  transferBytes[20] = SWIPE_Y_CONS_DIST_0;
  transferBytes[21] = SWIPE_Y_CONS_DIST_1;
  transferBytes[22] = SWIPE_ANGLE;
  transferBytes[23] = SCROLL_ANGLE;
  transferBytes[24] = ZOOM_INIT_DIST_0;
  transferBytes[25] = ZOOM_INIT_DIST_1;
  transferBytes[26] = ZOOM_CONSECUTIVE_DIST_0;
  transferBytes[27] = ZOOM_CONSECUTIVE_DIST_1;
  transferBytes[28] = SCROLL_INIT_DIST_0;
  transferBytes[29] = SCROLL_INIT_DIST_1;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_GESTURE_ENABLE_0, 30, transferBytes, RESTART);
  transferBytes[0] = SCROLL_CONSECUTIVE_DIST_0;
  transferBytes[1] = SCROLL_CONSECUTIVE_DIST_1;
  transferBytes[2] = PALM_GESTURE_THRESHOLD_0;
  transferBytes[3] = PALM_GESTURE_THRESHOLD_1;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_GESTURE_ENABLE_0+30, 4, transferBytes, RESTART);

  /* Change the Rx/Tx Mapping */
  /* Memory Map Position 0x1218 - 0x1245 */
  transferBytes[0] = RX_TX_MAP_0;
  transferBytes[1] = RX_TX_MAP_1;
  transferBytes[2] = RX_TX_MAP_2;
  transferBytes[3] = RX_TX_MAP_3;
  transferBytes[4] = RX_TX_MAP_4;
  transferBytes[5] = RX_TX_MAP_5;
  transferBytes[6] = RX_TX_MAP_6;
  transferBytes[7] = RX_TX_MAP_7;
  transferBytes[8] = RX_TX_MAP_8;
  transferBytes[9] = RX_TX_MAP_9;
  transferBytes[10] = RX_TX_MAP_10;
  transferBytes[11] = RX_TX_MAP_11;
  transferBytes[12] = RX_TX_MAP_12;
  transferBytes[13] = RX_TX_MAP_13;
  transferBytes[14] = RX_TX_MAP_14;
  transferBytes[15] = RX_TX_MAP_15;
  transferBytes[16] = RX_TX_MAP_16;
  transferBytes[17] = RX_TX_MAP_17;
  transferBytes[18] = RX_TX_MAP_18;
  transferBytes[19] = RX_TX_MAP_19;
  transferBytes[20] = RX_TX_MAP_20;
  transferBytes[21] = RX_TX_MAP_21;
  transferBytes[22] = RX_TX_MAP_22;
  transferBytes[23] = RX_TX_MAP_23;
  transferBytes[24] = RX_TX_MAP_24;
  transferBytes[25] = RX_TX_MAP_25;
  transferBytes[26] = RX_TX_MAP_26;
  transferBytes[27] = RX_TX_MAP_27;
  transferBytes[28] = RX_TX_MAP_28;
  transferBytes[29] = RX_TX_MAP_29;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_RX_TX_MAP_0, 30, transferBytes, RESTART);
  transferBytes[0] = RX_TX_MAP_30;
  transferBytes[1] = RX_TX_MAP_31;
  transferBytes[2] = RX_TX_MAP_32;
  transferBytes[3] = RX_TX_MAP_33;
  transferBytes[4] = RX_TX_MAP_34;
  transferBytes[5] = RX_TX_MAP_35;
  transferBytes[6] = RX_TX_MAP_36;
  transferBytes[7] = RX_TX_MAP_37;
  transferBytes[8] = RX_TX_MAP_38;
  transferBytes[9] = RX_TX_MAP_39;
  transferBytes[10] = RX_TX_MAP_40;
  transferBytes[11] = RX_TX_MAP_41;
  transferBytes[12] = RX_TX_MAP_42;
  transferBytes[13] = RX_TX_MAP_43;
  transferBytes[14] = RX_TX_MAP_44;
  transferBytes[15] = RX_TX_OPEN;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_RX_TX_MAP_0+30, 16, transferBytes, RESTART);

  /* Change the TP Channel Disables 0 */
  /* Memory Map Position 0x1246 - 0x1271 */
  transferBytes[0] = TP_DISABLE_TX0_0;
  transferBytes[1] = TP_DISABLE_TX0_1;
  transferBytes[2] = TP_DISABLE_TX0_2;
  transferBytes[3] = TP_DISABLE_TX0_3;
  transferBytes[4] = TP_DISABLE_TX1_0;
  transferBytes[5] = TP_DISABLE_TX1_1;
  transferBytes[6] = TP_DISABLE_TX1_2;
  transferBytes[7] = TP_DISABLE_TX1_3;
  transferBytes[8] = TP_DISABLE_TX2_0;
  transferBytes[9] = TP_DISABLE_TX2_1;
  transferBytes[10] = TP_DISABLE_TX2_2;
  transferBytes[11] = TP_DISABLE_TX2_3;
  transferBytes[12] = TP_DISABLE_TX3_0;
  transferBytes[13] = TP_DISABLE_TX3_1;
  transferBytes[14] = TP_DISABLE_TX3_2;
  transferBytes[15] = TP_DISABLE_TX3_3;
  transferBytes[16] = TP_DISABLE_TX4_0;
  transferBytes[17] = TP_DISABLE_TX4_1;
  transferBytes[18] = TP_DISABLE_TX4_2;
  transferBytes[19] = TP_DISABLE_TX4_3;
  transferBytes[20] = TP_DISABLE_TX5_0;
  transferBytes[21] = TP_DISABLE_TX5_1;
  transferBytes[22] = TP_DISABLE_TX5_2;
  transferBytes[23] = TP_DISABLE_TX5_3;
  transferBytes[24] = TP_DISABLE_TX6_0;
  transferBytes[25] = TP_DISABLE_TX6_1;
  transferBytes[26] = TP_DISABLE_TX6_2;
  transferBytes[27] = TP_DISABLE_TX6_3;
  transferBytes[28] = TP_DISABLE_TX7_0;
  transferBytes[29] = TP_DISABLE_TX7_1;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_TP_DISABLE_TX0_0, 30, transferBytes, RESTART);
  transferBytes[0] = TP_DISABLE_TX7_2;
  transferBytes[1] = TP_DISABLE_TX7_3;
  transferBytes[2] = TP_DISABLE_TX8_0;
  transferBytes[3] = TP_DISABLE_TX8_1;
  transferBytes[4] = TP_DISABLE_TX8_2;
  transferBytes[5] = TP_DISABLE_TX8_3;
  transferBytes[6] = TP_DISABLE_TX9_0;
  transferBytes[7] = TP_DISABLE_TX9_1;
  transferBytes[8] = TP_DISABLE_TX9_2;
  transferBytes[9] = TP_DISABLE_TX9_3;
  transferBytes[10] = TP_DISABLE_TX10_0;
  transferBytes[11] = TP_DISABLE_TX10_1;
  transferBytes[12] = TP_DISABLE_TX10_2;
  transferBytes[13] = TP_DISABLE_TX10_3;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_TP_DISABLE_TX0_0+30, 14, transferBytes, RESTART);

  /* Change the TP Channel Disables 1 */
  /* Memory Map Position 0x1272 - 0x129D */
  transferBytes[0] = TP_DISABLE_TX11_0;
  transferBytes[1] = TP_DISABLE_TX11_1;
  transferBytes[2] = TP_DISABLE_TX11_2;
  transferBytes[3] = TP_DISABLE_TX11_3;
  transferBytes[4] = TP_DISABLE_TX12_0;
  transferBytes[5] = TP_DISABLE_TX12_1;
  transferBytes[6] = TP_DISABLE_TX12_2;
  transferBytes[7] = TP_DISABLE_TX12_3;
  transferBytes[8] = TP_DISABLE_TX13_0;
  transferBytes[9] = TP_DISABLE_TX13_1;
  transferBytes[10] = TP_DISABLE_TX13_2;
  transferBytes[11] = TP_DISABLE_TX13_3;
  transferBytes[12] = TP_DISABLE_TX14_0;
  transferBytes[13] = TP_DISABLE_TX14_1;
  transferBytes[14] = TP_DISABLE_TX14_2;
  transferBytes[15] = TP_DISABLE_TX14_3;
  transferBytes[16] = TP_DISABLE_TX15_0;
  transferBytes[17] = TP_DISABLE_TX15_1;
  transferBytes[18] = TP_DISABLE_TX15_2;
  transferBytes[19] = TP_DISABLE_TX15_3;
  transferBytes[20] = TP_DISABLE_TX16_0;
  transferBytes[21] = TP_DISABLE_TX16_1;
  transferBytes[22] = TP_DISABLE_TX16_2;
  transferBytes[23] = TP_DISABLE_TX16_3;
  transferBytes[24] = TP_DISABLE_TX17_0;
  transferBytes[25] = TP_DISABLE_TX17_1;
  transferBytes[26] = TP_DISABLE_TX17_2;
  transferBytes[27] = TP_DISABLE_TX17_3;
  transferBytes[28] = TP_DISABLE_TX18_0;
  transferBytes[29] = TP_DISABLE_TX18_1;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_TP_DISABLE_TX11_0, 30, transferBytes, RESTART);
  transferBytes[0] = TP_DISABLE_TX18_2;
  transferBytes[1] = TP_DISABLE_TX18_3;
  transferBytes[2] = TP_DISABLE_TX19_0;
  transferBytes[3] = TP_DISABLE_TX19_1;
  transferBytes[4] = TP_DISABLE_TX19_2;
  transferBytes[5] = TP_DISABLE_TX19_3;
  transferBytes[6] = TP_DISABLE_TX20_0;
  transferBytes[7] = TP_DISABLE_TX20_1;
  transferBytes[8] = TP_DISABLE_TX20_2;
  transferBytes[9] = TP_DISABLE_TX20_3;
  transferBytes[10] = TP_DISABLE_TX21_0;
  transferBytes[11] = TP_DISABLE_TX21_1;
  transferBytes[12] = TP_DISABLE_TX21_2;
  transferBytes[13] = TP_DISABLE_TX21_3;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_TP_DISABLE_TX11_0+30, 14, transferBytes, RESTART);

  /* Change the TP Snap Enable 0 */
  /* Memory Map Position 0x129E - 0x12C9 */
  transferBytes[0] = SNAP_ENABLE_TX0_0;
  transferBytes[1] = SNAP_ENABLE_TX0_1;
  transferBytes[2] = SNAP_ENABLE_TX0_2;
  transferBytes[3] = SNAP_ENABLE_TX0_3;
  transferBytes[4] = SNAP_ENABLE_TX1_0;
  transferBytes[5] = SNAP_ENABLE_TX1_1;
  transferBytes[6] = SNAP_ENABLE_TX1_2;
  transferBytes[7] = SNAP_ENABLE_TX1_3;
  transferBytes[8] = SNAP_ENABLE_TX2_0;
  transferBytes[9] = SNAP_ENABLE_TX2_1;
  transferBytes[10] = SNAP_ENABLE_TX2_2;
  transferBytes[11] = SNAP_ENABLE_TX2_3;
  transferBytes[12] = SNAP_ENABLE_TX3_0;
  transferBytes[13] = SNAP_ENABLE_TX3_1;
  transferBytes[14] = SNAP_ENABLE_TX3_2;
  transferBytes[15] = SNAP_ENABLE_TX3_3;
  transferBytes[16] = SNAP_ENABLE_TX4_0;
  transferBytes[17] = SNAP_ENABLE_TX4_1;
  transferBytes[18] = SNAP_ENABLE_TX4_2;
  transferBytes[19] = SNAP_ENABLE_TX4_3;
  transferBytes[20] = SNAP_ENABLE_TX5_0;
  transferBytes[21] = SNAP_ENABLE_TX5_1;
  transferBytes[22] = SNAP_ENABLE_TX5_2;
  transferBytes[23] = SNAP_ENABLE_TX5_3;
  transferBytes[24] = SNAP_ENABLE_TX6_0;
  transferBytes[25] = SNAP_ENABLE_TX6_1;
  transferBytes[26] = SNAP_ENABLE_TX6_2;
  transferBytes[27] = SNAP_ENABLE_TX6_3;
  transferBytes[28] = SNAP_ENABLE_TX7_0;
  transferBytes[29] = SNAP_ENABLE_TX7_1;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_SNAP_ENABLE_TX0_0, 30, transferBytes, RESTART);
  transferBytes[0] = SNAP_ENABLE_TX7_2;
  transferBytes[1] = SNAP_ENABLE_TX7_3;
  transferBytes[2] = SNAP_ENABLE_TX8_0;
  transferBytes[3] = SNAP_ENABLE_TX8_1;
  transferBytes[4] = SNAP_ENABLE_TX8_2;
  transferBytes[5] = SNAP_ENABLE_TX8_3;
  transferBytes[6] = SNAP_ENABLE_TX9_0;
  transferBytes[7] = SNAP_ENABLE_TX9_1;
  transferBytes[8] = SNAP_ENABLE_TX9_2;
  transferBytes[9] = SNAP_ENABLE_TX9_3;
  transferBytes[10] = SNAP_ENABLE_TX10_0;
  transferBytes[11] = SNAP_ENABLE_TX10_1;
  transferBytes[12] = SNAP_ENABLE_TX10_2;
  transferBytes[13] = SNAP_ENABLE_TX10_3;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_SNAP_ENABLE_TX0_0+30, 14, transferBytes, RESTART);

  /* Change the TP Snap Enable 1 */
  /* Memory Map Position 0x12CA - 0x12F5 */
  transferBytes[0] = SNAP_ENABLE_TX11_0;
  transferBytes[1] = SNAP_ENABLE_TX11_1;
  transferBytes[2] = SNAP_ENABLE_TX11_2;
  transferBytes[3] = SNAP_ENABLE_TX11_3;
  transferBytes[4] = SNAP_ENABLE_TX12_0;
  transferBytes[5] = SNAP_ENABLE_TX12_1;
  transferBytes[6] = SNAP_ENABLE_TX12_2;
  transferBytes[7] = SNAP_ENABLE_TX12_3;
  transferBytes[8] = SNAP_ENABLE_TX13_0;
  transferBytes[9] = SNAP_ENABLE_TX13_1;
  transferBytes[10] = SNAP_ENABLE_TX13_2;
  transferBytes[11] = SNAP_ENABLE_TX13_3;
  transferBytes[12] = SNAP_ENABLE_TX14_0;
  transferBytes[13] = SNAP_ENABLE_TX14_1;
  transferBytes[14] = SNAP_ENABLE_TX14_2;
  transferBytes[15] = SNAP_ENABLE_TX14_3;
  transferBytes[16] = SNAP_ENABLE_TX15_0;
  transferBytes[17] = SNAP_ENABLE_TX15_1;
  transferBytes[18] = SNAP_ENABLE_TX15_2;
  transferBytes[19] = SNAP_ENABLE_TX15_3;
  transferBytes[20] = SNAP_ENABLE_TX16_0;
  transferBytes[21] = SNAP_ENABLE_TX16_1;
  transferBytes[22] = SNAP_ENABLE_TX16_2;
  transferBytes[23] = SNAP_ENABLE_TX16_3;
  transferBytes[24] = SNAP_ENABLE_TX17_0;
  transferBytes[25] = SNAP_ENABLE_TX17_1;
  transferBytes[26] = SNAP_ENABLE_TX17_2;
  transferBytes[27] = SNAP_ENABLE_TX17_3;
  transferBytes[28] = SNAP_ENABLE_TX18_0;
  transferBytes[29] = SNAP_ENABLE_TX18_1;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_SNAP_ENABLE_TX11_0, 30, transferBytes, RESTART);
  transferBytes[0] = SNAP_ENABLE_TX18_2;
  transferBytes[1] = SNAP_ENABLE_TX18_3;
  transferBytes[2] = SNAP_ENABLE_TX19_0;
  transferBytes[3] = SNAP_ENABLE_TX19_1;
  transferBytes[4] = SNAP_ENABLE_TX19_2;
  transferBytes[5] = SNAP_ENABLE_TX19_3;
  transferBytes[6] = SNAP_ENABLE_TX20_0;
  transferBytes[7] = SNAP_ENABLE_TX20_1;
  transferBytes[8] = SNAP_ENABLE_TX20_2;
  transferBytes[9] = SNAP_ENABLE_TX20_3;
  transferBytes[10] = SNAP_ENABLE_TX21_0;
  transferBytes[11] = SNAP_ENABLE_TX21_1;
  transferBytes[12] = SNAP_ENABLE_TX21_2;
  transferBytes[13] = SNAP_ENABLE_TX21_3;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_SNAP_ENABLE_TX11_0+30, 14, transferBytes, RESTART);

  /* Change the Individual Touch Threshold Adjustments */
  /* Memory Map Position 0x12F6 - 0x12FF */
  transferBytes[0] = TT_ADJUST_CH0;
  transferBytes[1] = TT_ADJUST_CH1;
  transferBytes[2] = TT_ADJUST_CH2;
  transferBytes[3] = TT_ADJUST_CH3;
  transferBytes[4] = TT_ADJUST_CH4;
  transferBytes[5] = TT_ADJUST_CH5;
  transferBytes[6] = TT_ADJUST_CH6;
  transferBytes[7] = TT_ADJUST_CH7;
  transferBytes[8] = TT_ADJUST_CH8;
  transferBytes[9] = TT_ADJUST_CH9;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_TT_ADJUST_CH0, 10, transferBytes, RESTART);

  /* Change the Virtual Buttons 0 */
  /* Memory Map Position 0x14F0 - 0x1511 */
  transferBytes[0] = NUMBER_OF_VIRTUAL_SENSORS_0;
  transferBytes[1] = NUMBER_OF_VIRTUAL_SENSORS_1;
  transferBytes[2] = B0_TOP_LEFT_X_0;
  transferBytes[3] = B0_TOP_LEFT_X_1;
  transferBytes[4] = B0_TOP_LEFT_Y_0;
  transferBytes[5] = B0_TOP_LEFT_Y_1;
  transferBytes[6] = B0_BOT_RIGHT_X_0;
  transferBytes[7] = B0_BOT_RIGHT_X_1;
  transferBytes[8] = B0_BOT_RIGHT_Y_0;
  transferBytes[9] = B0_BOT_RIGHT_Y_1;
  transferBytes[10] = B1_TOP_LEFT_X_0;
  transferBytes[11] = B1_TOP_LEFT_X_1;
  transferBytes[12] = B1_TOP_LEFT_Y_0;
  transferBytes[13] = B1_TOP_LEFT_Y_1;
  transferBytes[14] = B1_BOT_RIGHT_X_0;
  transferBytes[15] = B1_BOT_RIGHT_X_1;
  transferBytes[16] = B1_BOT_RIGHT_Y_0;
  transferBytes[17] = B1_BOT_RIGHT_Y_1;
  transferBytes[18] = B2_TOP_LEFT_X_0;
  transferBytes[19] = B2_TOP_LEFT_X_1;
  transferBytes[20] = B2_TOP_LEFT_Y_0;
  transferBytes[21] = B2_TOP_LEFT_Y_1;
  transferBytes[22] = B2_BOT_RIGHT_X_0;
  transferBytes[23] = B2_BOT_RIGHT_X_1;
  transferBytes[24] = B2_BOT_RIGHT_Y_0;
  transferBytes[25] = B2_BOT_RIGHT_Y_1;
  transferBytes[26] = B3_TOP_LEFT_X_0;
  transferBytes[27] = B3_TOP_LEFT_X_1;
  transferBytes[28] = B3_TOP_LEFT_Y_0;
  transferBytes[29] = B3_TOP_LEFT_Y_1;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_NUMBER_OF_VIRTUAL_SENSORS_0, 30, transferBytes, RESTART);
  transferBytes[0] = B3_BOT_RIGHT_X_0;
  transferBytes[1] = B3_BOT_RIGHT_X_1;
  transferBytes[2] = B3_BOT_RIGHT_Y_0;
  transferBytes[3] = B3_BOT_RIGHT_Y_1;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_NUMBER_OF_VIRTUAL_SENSORS_0+30, 4, transferBytes, RESTART);

  /* Change the Virtual Buttons 1 */
  /* Memory Map Position 0x1512 - 0x1541 */
  transferBytes[0] = B4_TOP_LEFT_X_0;
  transferBytes[1] = B4_TOP_LEFT_X_1;
  transferBytes[2] = B4_TOP_LEFT_Y_0;
  transferBytes[3] = B4_TOP_LEFT_Y_1;
  transferBytes[4] = B4_BOT_RIGHT_X_0;
  transferBytes[5] = B4_BOT_RIGHT_X_1;
  transferBytes[6] = B4_BOT_RIGHT_Y_0;
  transferBytes[7] = B4_BOT_RIGHT_Y_1;
  transferBytes[8] = B5_TOP_LEFT_X_0;
  transferBytes[9] = B5_TOP_LEFT_X_1;
  transferBytes[10] = B5_TOP_LEFT_Y_0;
  transferBytes[11] = B5_TOP_LEFT_Y_1;
  transferBytes[12] = B5_BOT_RIGHT_X_0;
  transferBytes[13] = B5_BOT_RIGHT_X_1;
  transferBytes[14] = B5_BOT_RIGHT_Y_0;
  transferBytes[15] = B5_BOT_RIGHT_Y_1;
  transferBytes[16] = B6_TOP_LEFT_X_0;
  transferBytes[17] = B6_TOP_LEFT_X_1;
  transferBytes[18] = B6_TOP_LEFT_Y_0;
  transferBytes[19] = B6_TOP_LEFT_Y_1;
  transferBytes[20] = B6_BOT_RIGHT_X_0;
  transferBytes[21] = B6_BOT_RIGHT_X_1;
  transferBytes[22] = B6_BOT_RIGHT_Y_0;
  transferBytes[23] = B6_BOT_RIGHT_Y_1;
  transferBytes[24] = B7_TOP_LEFT_X_0;
  transferBytes[25] = B7_TOP_LEFT_X_1;
  transferBytes[26] = B7_TOP_LEFT_Y_0;
  transferBytes[27] = B7_TOP_LEFT_Y_1;
  transferBytes[28] = B7_BOT_RIGHT_X_0;
  transferBytes[29] = B7_BOT_RIGHT_X_1;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_B4_TOP_LEFT_X_0, 30, transferBytes, RESTART);
  transferBytes[0] = B7_BOT_RIGHT_Y_0;
  transferBytes[1] = B7_BOT_RIGHT_Y_1;
  transferBytes[2] = B8_TOP_LEFT_X_0;
  transferBytes[3] = B8_TOP_LEFT_X_1;
  transferBytes[4] = B8_TOP_LEFT_Y_0;
  transferBytes[5] = B8_TOP_LEFT_Y_1;
  transferBytes[6] = B8_BOT_RIGHT_X_0;
  transferBytes[7] = B8_BOT_RIGHT_X_1;
  transferBytes[8] = B8_BOT_RIGHT_Y_0;
  transferBytes[9] = B8_BOT_RIGHT_Y_1;
  transferBytes[10] = B9_TOP_LEFT_X_0;
  transferBytes[11] = B9_TOP_LEFT_X_1;
  transferBytes[12] = B9_TOP_LEFT_Y_0;
  transferBytes[13] = B9_TOP_LEFT_Y_1;
  transferBytes[14] = B9_BOT_RIGHT_X_0;
  transferBytes[15] = B9_BOT_RIGHT_X_1;
  transferBytes[16] = B9_BOT_RIGHT_Y_0;
  transferBytes[17] = B9_BOT_RIGHT_Y_1;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_B4_TOP_LEFT_X_0+30, 18, transferBytes, RESTART);

  /* Change the Virtual Buttons 2 */
  /* Memory Map Position 0x1542 - 0x1571 */
  transferBytes[0] = B10_TOP_LEFT_X_0;
  transferBytes[1] = B10_TOP_LEFT_X_1;
  transferBytes[2] = B10_TOP_LEFT_Y_0;
  transferBytes[3] = B10_TOP_LEFT_Y_1;
  transferBytes[4] = B10_BOT_RIGHT_X_0;
  transferBytes[5] = B10_BOT_RIGHT_X_1;
  transferBytes[6] = B10_BOT_RIGHT_Y_0;
  transferBytes[7] = B10_BOT_RIGHT_Y_1;
  transferBytes[8] = B11_TOP_LEFT_X_0;
  transferBytes[9] = B11_TOP_LEFT_X_1;
  transferBytes[10] = B11_TOP_LEFT_Y_0;
  transferBytes[11] = B11_TOP_LEFT_Y_1;
  transferBytes[12] = B11_BOT_RIGHT_X_0;
  transferBytes[13] = B11_BOT_RIGHT_X_1;
  transferBytes[14] = B11_BOT_RIGHT_Y_0;
  transferBytes[15] = B11_BOT_RIGHT_Y_1;
  transferBytes[16] = B12_TOP_LEFT_X_0;
  transferBytes[17] = B12_TOP_LEFT_X_1;
  transferBytes[18] = B12_TOP_LEFT_Y_0;
  transferBytes[19] = B12_TOP_LEFT_Y_1;
  transferBytes[20] = B12_BOT_RIGHT_X_0;
  transferBytes[21] = B12_BOT_RIGHT_X_1;
  transferBytes[22] = B12_BOT_RIGHT_Y_0;
  transferBytes[23] = B12_BOT_RIGHT_Y_1;
  transferBytes[24] = B13_TOP_LEFT_X_0;
  transferBytes[25] = B13_TOP_LEFT_X_1;
  transferBytes[26] = B13_TOP_LEFT_Y_0;
  transferBytes[27] = B13_TOP_LEFT_Y_1;
  transferBytes[28] = B13_BOT_RIGHT_X_0;
  transferBytes[29] = B13_BOT_RIGHT_X_1;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_B10_TOP_LEFT_X_0, 30, transferBytes, RESTART);
  transferBytes[0] = B13_BOT_RIGHT_Y_0;
  transferBytes[1] = B13_BOT_RIGHT_Y_1;
  transferBytes[2] = B14_TOP_LEFT_X_0;
  transferBytes[3] = B14_TOP_LEFT_X_1;
  transferBytes[4] = B14_TOP_LEFT_Y_0;
  transferBytes[5] = B14_TOP_LEFT_Y_1;
  transferBytes[6] = B14_BOT_RIGHT_X_0;
  transferBytes[7] = B14_BOT_RIGHT_X_1;
  transferBytes[8] = B14_BOT_RIGHT_Y_0;
  transferBytes[9] = B14_BOT_RIGHT_Y_1;
  transferBytes[10] = B15_TOP_LEFT_X_0;
  transferBytes[11] = B15_TOP_LEFT_X_1;
  transferBytes[12] = B15_TOP_LEFT_Y_0;
  transferBytes[13] = B15_TOP_LEFT_Y_1;
  transferBytes[14] = B15_BOT_RIGHT_X_0;
  transferBytes[15] = B15_BOT_RIGHT_X_1;
  transferBytes[16] = B15_BOT_RIGHT_Y_0;
  transferBytes[17] = B15_BOT_RIGHT_Y_1;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_B10_TOP_LEFT_X_0+30, 18, transferBytes, RESTART);

  /* Change the Virtual Sliders 0 */
  /* Memory Map Position 0x1572 - 0x15A5 */
  transferBytes[0] = SLIDER_DEADZONE_0;
  transferBytes[1] = SLIDER_DEADZONE_1;
  transferBytes[2] = S0_TOP_LEFT_X_0;
  transferBytes[3] = S0_TOP_LEFT_X_1;
  transferBytes[4] = S0_TOP_LEFT_Y_0;
  transferBytes[5] = S0_TOP_LEFT_Y_1;
  transferBytes[6] = S0_BOT_RIGHT_X_0;
  transferBytes[7] = S0_BOT_RIGHT_X_1;
  transferBytes[8] = S0_BOT_RIGHT_Y_0;
  transferBytes[9] = S0_BOT_RIGHT_Y_1;
  transferBytes[10] = S0_RESOLUTION_0;
  transferBytes[11] = S0_RESOLUTION_1;
  transferBytes[12] = S1_TOP_LEFT_X_0;
  transferBytes[13] = S1_TOP_LEFT_X_1;
  transferBytes[14] = S1_TOP_LEFT_Y_0;
  transferBytes[15] = S1_TOP_LEFT_Y_1;
  transferBytes[16] = S1_BOT_RIGHT_X_0;
  transferBytes[17] = S1_BOT_RIGHT_X_1;
  transferBytes[18] = S1_BOT_RIGHT_Y_0;
  transferBytes[19] = S1_BOT_RIGHT_Y_1;
  transferBytes[20] = S1_RESOLUTION_0;
  transferBytes[21] = S1_RESOLUTION_1;
  transferBytes[22] = S2_TOP_LEFT_X_0;
  transferBytes[23] = S2_TOP_LEFT_X_1;
  transferBytes[24] = S2_TOP_LEFT_Y_0;
  transferBytes[25] = S2_TOP_LEFT_Y_1;
  transferBytes[26] = S2_BOT_RIGHT_X_0;
  transferBytes[27] = S2_BOT_RIGHT_X_1;
  transferBytes[28] = S2_BOT_RIGHT_Y_0;
  transferBytes[29] = S2_BOT_RIGHT_Y_1;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_SLIDER_DEADZONE_0, 30, transferBytes, RESTART);
  transferBytes[0] = S2_RESOLUTION_0;
  transferBytes[1] = S2_RESOLUTION_1;
  transferBytes[2] = S3_TOP_LEFT_X_0;
  transferBytes[3] = S3_TOP_LEFT_X_1;
  transferBytes[4] = S3_TOP_LEFT_Y_0;
  transferBytes[5] = S3_TOP_LEFT_Y_1;
  transferBytes[6] = S3_BOT_RIGHT_X_0;
  transferBytes[7] = S3_BOT_RIGHT_X_1;
  transferBytes[8] = S3_BOT_RIGHT_Y_0;
  transferBytes[9] = S3_BOT_RIGHT_Y_1;
  transferBytes[10] = S3_RESOLUTION_0;
  transferBytes[11] = S3_RESOLUTION_1;
  transferBytes[12] = S4_TOP_LEFT_X_0;
  transferBytes[13] = S4_TOP_LEFT_X_1;
  transferBytes[14] = S4_TOP_LEFT_Y_0;
  transferBytes[15] = S4_TOP_LEFT_Y_1;
  transferBytes[16] = S4_BOT_RIGHT_X_0;
  transferBytes[17] = S4_BOT_RIGHT_X_1;
  transferBytes[18] = S4_BOT_RIGHT_Y_0;
  transferBytes[19] = S4_BOT_RIGHT_Y_1;
  transferBytes[20] = S4_RESOLUTION_0;
  transferBytes[21] = S4_RESOLUTION_1;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_SLIDER_DEADZONE_0+30, 22, transferBytes, RESTART);

  /* Change the Virtual Sliders 1 */
  /* Memory Map Position 0x15A6 - 0x15C3 */
  transferBytes[0] = S5_TOP_LEFT_X_0;
  transferBytes[1] = S5_TOP_LEFT_X_1;
  transferBytes[2] = S5_TOP_LEFT_Y_0;
  transferBytes[3] = S5_TOP_LEFT_Y_1;
  transferBytes[4] = S5_BOT_RIGHT_X_0;
  transferBytes[5] = S5_BOT_RIGHT_X_1;
  transferBytes[6] = S5_BOT_RIGHT_Y_0;
  transferBytes[7] = S5_BOT_RIGHT_Y_1;
  transferBytes[8] = S5_RESOLUTION_0;
  transferBytes[9] = S5_RESOLUTION_1;
  transferBytes[10] = S6_TOP_LEFT_X_0;
  transferBytes[11] = S6_TOP_LEFT_X_1;
  transferBytes[12] = S6_TOP_LEFT_Y_0;
  transferBytes[13] = S6_TOP_LEFT_Y_1;
  transferBytes[14] = S6_BOT_RIGHT_X_0;
  transferBytes[15] = S6_BOT_RIGHT_X_1;
  transferBytes[16] = S6_BOT_RIGHT_Y_0;
  transferBytes[17] = S6_BOT_RIGHT_Y_1;
  transferBytes[18] = S6_RESOLUTION_0;
  transferBytes[19] = S6_RESOLUTION_1;
  transferBytes[20] = S7_TOP_LEFT_X_0;
  transferBytes[21] = S7_TOP_LEFT_X_1;
  transferBytes[22] = S7_TOP_LEFT_Y_0;
  transferBytes[23] = S7_TOP_LEFT_Y_1;
  transferBytes[24] = S7_BOT_RIGHT_X_0;
  transferBytes[25] = S7_BOT_RIGHT_X_1;
  transferBytes[26] = S7_BOT_RIGHT_Y_0;
  transferBytes[27] = S7_BOT_RIGHT_Y_1;
  transferBytes[28] = S7_RESOLUTION_0;
  transferBytes[29] = S7_RESOLUTION_1;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_S5_TOP_LEFT_X_0, 30, transferBytes, RESTART);

  /* Change the Virtual Wheels */
  /* Memory Map Position 0x15C4 - 0x15EB */
  transferBytes[0] = W0_CENTRE_X_0;
  transferBytes[1] = W0_CENTRE_X_1;
  transferBytes[2] = W0_CENTRE_Y_0;
  transferBytes[3] = W0_CENTRE_Y_1;
  transferBytes[4] = W0_INNER_RADIUS_0;
  transferBytes[5] = W0_INNER_RADIUS_1;
  transferBytes[6] = W0_OUTER_RADIUS_0;
  transferBytes[7] = W0_OUTER_RADIUS_1;
  transferBytes[8] = W0_RESOLUTION_0;
  transferBytes[9] = W0_RESOLUTION_1;
  transferBytes[10] = W1_CENTRE_X_0;
  transferBytes[11] = W1_CENTRE_X_1;
  transferBytes[12] = W1_CENTRE_Y_0;
  transferBytes[13] = W1_CENTRE_Y_1;
  transferBytes[14] = W1_INNER_RADIUS_0;
  transferBytes[15] = W1_INNER_RADIUS_1;
  transferBytes[16] = W1_OUTER_RADIUS_0;
  transferBytes[17] = W1_OUTER_RADIUS_1;
  transferBytes[18] = W1_RESOLUTION_0;
  transferBytes[19] = W1_RESOLUTION_1;
  transferBytes[20] = W2_CENTRE_X_0;
  transferBytes[21] = W2_CENTRE_X_1;
  transferBytes[22] = W2_CENTRE_Y_0;
  transferBytes[23] = W2_CENTRE_Y_1;
  transferBytes[24] = W2_INNER_RADIUS_0;
  transferBytes[25] = W2_INNER_RADIUS_1;
  transferBytes[26] = W2_OUTER_RADIUS_0;
  transferBytes[27] = W2_OUTER_RADIUS_1;
  transferBytes[28] = W2_RESOLUTION_0;
  transferBytes[29] = W2_RESOLUTION_1;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_W0_CENTRE_X_0, 30, transferBytes, RESTART);
  transferBytes[0] = W3_CENTRE_X_0;
  transferBytes[1] = W3_CENTRE_X_1;
  transferBytes[2] = W3_CENTRE_Y_0;
  transferBytes[3] = W3_CENTRE_Y_1;
  transferBytes[4] = W3_INNER_RADIUS_0;
  transferBytes[5] = W3_INNER_RADIUS_1;
  transferBytes[6] = W3_OUTER_RADIUS_0;
  transferBytes[7] = W3_OUTER_RADIUS_1;
  transferBytes[8] = W3_RESOLUTION_0;
  transferBytes[9] = W3_RESOLUTION_1;
  writeRandomBytes16(_deviceAddress, IQS9150_MM_W0_CENTRE_X_0+30, 10, transferBytes, RESTART);
}

/**
  * @name   getPowerMode
  * @brief  A method that returns the current power mode state of the IQS9150.
  * @param  None
  * @retval iqs9150_power_mode_e -> power mode enum, Active, idle-touch, idle, lp1 or lp2.
*/
iqs9150_power_mode_e IQS9150::getPowerMode(void)
{
  uint8_t powerMode = IQSMemoryMap.INFO_FLAGS[0]; // load the system status byte
  powerMode = powerMode & 0x07; // extract the power mode bits
  return (iqs9150_power_mode_e)powerMode;
}

/**
 * @name	getAbsXCoordinate
 * @brief  A method that returns the constructed 16-bit coordinate value
 * @param  fingerNum     -> Specifies the finger number. Finger 1 is the first
 *                          finger to touch the trackpad, finger 2 is the second
 *                          to touch the trackpad.
 * @retval Returns 16-bit X coordinate value.
 */
uint16_t IQS9150::getAbsXCoordinate(uint8_t fingerNum)
{
  /* The 16-bit return value. */
  uint16_t absXCoordReturn = 0;

  /* Construct the 16-bit return value. */
  switch (fingerNum)
  {
  case 1:
    absXCoordReturn = (uint16_t)(IQSMemoryMap.FINGER_1_X[0]);
    absXCoordReturn |= (uint16_t)(IQSMemoryMap.FINGER_1_X[1] << 8);
    break;
  case 2:
    absXCoordReturn = (uint16_t)(IQSMemoryMap.FINGER_2_X[0]);
    absXCoordReturn |= (uint16_t)(IQSMemoryMap.FINGER_2_X[1] << 8);
    break;
  case 3:
    absXCoordReturn = (uint16_t)(IQSMemoryMap.FINGER_3_X[0]);
    absXCoordReturn |= (uint16_t)(IQSMemoryMap.FINGER_3_X[1] << 8);
    break;
  case 4:
    absXCoordReturn = (uint16_t)(IQSMemoryMap.FINGER_4_X[0]);
    absXCoordReturn |= (uint16_t)(IQSMemoryMap.FINGER_4_X[1] << 8);
    break;
  case 5:
    absXCoordReturn = (uint16_t)(IQSMemoryMap.FINGER_5_X[0]);
    absXCoordReturn |= (uint16_t)(IQSMemoryMap.FINGER_5_X[1] << 8);
    break;
  case 6:
    absXCoordReturn = (uint16_t)(IQSMemoryMap.FINGER_6_X[0]);
    absXCoordReturn |= (uint16_t)(IQSMemoryMap.FINGER_6_X[1] << 8);
    break;
  case 7:
    absXCoordReturn = (uint16_t)(IQSMemoryMap.FINGER_7_X[0]);
    absXCoordReturn |= (uint16_t)(IQSMemoryMap.FINGER_7_X[1] << 8);
    break;

  default:
    return 0xFFFF;
    break;
  }

  /*- Return the coordinate value.
    - Note that a value of 65535 (0xFFFF) means there is no touch. */
  return absXCoordReturn;
}

/**
 * @name	 getAbsYCoordinate
 * @brief  A method that returns the constructed 16-bit coordinate value
 * @param  fingerNum     -> Specifies the finger number. Finger 1 is the first
 *                          finger to touch the trackpad, finger 2 is the second
 *                          to touch the trackpad.
 * @retval Returns 16-bit Y coordinate value.
 */
uint16_t IQS9150::getAbsYCoordinate(uint8_t fingerNum)
{
  /* The 16-bit return value. */
  uint16_t absYCoordReturn = 0;

  /* Construct the 16-bit return value. */
  switch (fingerNum)
  {
  case 1:
    absYCoordReturn = (uint16_t)(IQSMemoryMap.FINGER_1_Y[0]);
    absYCoordReturn |= (uint16_t)(IQSMemoryMap.FINGER_1_Y[1] << 8);
    break;
  case 2:
    absYCoordReturn = (uint16_t)(IQSMemoryMap.FINGER_2_Y[0]);
    absYCoordReturn |= (uint16_t)(IQSMemoryMap.FINGER_2_Y[1] << 8);
    break;
  case 3:
    absYCoordReturn = (uint16_t)(IQSMemoryMap.FINGER_3_Y[0]);
    absYCoordReturn |= (uint16_t)(IQSMemoryMap.FINGER_3_Y[1] << 8);
    break;
  case 4:
    absYCoordReturn = (uint16_t)(IQSMemoryMap.FINGER_4_Y[0]);
    absYCoordReturn |= (uint16_t)(IQSMemoryMap.FINGER_4_Y[1] << 8);
    break;
  case 5:
    absYCoordReturn = (uint16_t)(IQSMemoryMap.FINGER_5_Y[0]);
    absYCoordReturn |= (uint16_t)(IQSMemoryMap.FINGER_5_Y[1] << 8);
    break;
  case 6:
    absYCoordReturn = (uint16_t)(IQSMemoryMap.FINGER_6_Y[0]);
    absYCoordReturn |= (uint16_t)(IQSMemoryMap.FINGER_6_Y[1] << 8);
    break;
  case 7:
    absYCoordReturn = (uint16_t)(IQSMemoryMap.FINGER_7_Y[0]);
    absYCoordReturn |= (uint16_t)(IQSMemoryMap.FINGER_7_Y[1] << 8);
    break;

  default:
    return 0xFFFF;
    break;
  }

  /*- Return the coordinate value.
    - Note that a value of 65535 (0xFFFF) means there is no touch. */
  return absYCoordReturn;
}

/**
  * @name   getTouchpadEvent
  * @brief  A method that gives the type of event that occurred on the touchpad.
  * @retval Returns the slider event that has occurred defined as iqs9150_gestures_e.
  * @note   See the iqs9150_gestures_e typedef for all possible slider events
  *         that can occur.
  */
iqs9150_gestures_e IQS9150::getTouchpadEvent(void)
{

  /* Find the slider event that occurred */
  if(getBit(IQSMemoryMap.SINGLE_FINGER_GESTURES[0], IQS9150_GESTURE_SINGLE_TAP_BIT))
  {
    return IQS9150_SINGLE_GESTURE_SINGLE_TAP;
  }
  else if(getBit(IQSMemoryMap.SINGLE_FINGER_GESTURES[0], IQS9150_GESTURE_DOUBLE_TAP_BIT))
  {
    return IQS9150_SINGLE_GESTURE_DOUBLE_TAP;
  }
  else if(getBit(IQSMemoryMap.SINGLE_FINGER_GESTURES[0], IQS9150_GESTURE_TRIPLE_TAP_BIT))
  {
    return IQS9150_SINGLE_GESTURE_TRIPLE_TAP;
  }
  else if(getBit(IQSMemoryMap.SINGLE_FINGER_GESTURES[0], IQS9150_GESTURE_PRESS_HOLD_BIT))
  {
    return IQS9150_SINGLE_GESTURE_PRESS_HOLD;
  }
  else if(getBit(IQSMemoryMap.SINGLE_FINGER_GESTURES[0], IQS9150_SINGLE_GESTURE_PALM_GESTURE_BIT))
  {
    return IQS9150_SINGLE_GESTURE_PALM_GESTURE;
  }
  else if(getBit(IQSMemoryMap.SINGLE_FINGER_GESTURES[1], IQS9150_SINGLE_GESTURE_SWIPE_X_POSITIVE_BIT))
  {
    return IQS9150_SINGLE_GESTURE_SWIPE_X_POSITIVE;
  }
  else if(getBit(IQSMemoryMap.SINGLE_FINGER_GESTURES[1], IQS9150_SINGLE_GESTURE_SWIPE_X_NEGATIVE_BIT))
  {
    return IQS9150_SINGLE_GESTURE_SWIPE_X_NEGATIVE;
  }
  else if(getBit(IQSMemoryMap.SINGLE_FINGER_GESTURES[1], IQS9150_SINGLE_GESTURE_SWIPE_Y_POSITIVE_BIT))
  {
    return IQS9150_SINGLE_GESTURE_SWIPE_Y_POSITIVE;
  }
  else if(getBit(IQSMemoryMap.SINGLE_FINGER_GESTURES[1], IQS9150_SINGLE_GESTURE_SWIPE_Y_NEGATIVE_BIT))
  {
    return IQS9150_SINGLE_GESTURE_SWIPE_Y_NEGATIVE;
  }
  else if(getBit(IQSMemoryMap.SINGLE_FINGER_GESTURES[1], IQS9150_SINGLE_GESTURE_SWIPE_HOLD_X_POSITIVE_BIT))
  {
    return IQS9150_SINGLE_GESTURE_SWIPE_HOLD_X_POSITIVE;
  }
  else if(getBit(IQSMemoryMap.SINGLE_FINGER_GESTURES[1], IQS9150_SINGLE_GESTURE_SWIPE_HOLD_X_NEGATIVE_BIT))
  {
    return IQS9150_SINGLE_GESTURE_SWIPE_HOLD_X_NEGATIVE;
  }
  else if(getBit(IQSMemoryMap.SINGLE_FINGER_GESTURES[1], IQS9150_SINGLE_GESTURE_SWIPE_HOLD_Y_POSITIVE_BIT))
  {
    return IQS9150_SINGLE_GESTURE_SWIPE_HOLD_Y_POSITIVE;
  }
  else if(getBit(IQSMemoryMap.SINGLE_FINGER_GESTURES[1], IQS9150_SINGLE_GESTURE_SWIPE_HOLD_Y_NEGATIVE_BIT))
  {
    return IQS9150_SINGLE_GESTURE_SWIPE_HOLD_Y_NEGATIVE;
  }
  else if(getBit(IQSMemoryMap.DOUBLE_FINGER_GESTURES[0], IQS9150_GESTURE_SINGLE_TAP_BIT))
  {
    return IQS9150_DOUBLE_GESTURE_SINGLE_TAP;
  }
  else if(getBit(IQSMemoryMap.DOUBLE_FINGER_GESTURES[0], IQS9150_GESTURE_DOUBLE_TAP_BIT))
  {
    return IQS9150_DOUBLE_GESTURE_DOUBLE_TAP;
  }
  else if(getBit(IQSMemoryMap.DOUBLE_FINGER_GESTURES[0], IQS9150_GESTURE_TRIPLE_TAP_BIT))
  {
    return IQS9150_DOUBLE_GESTURE_TRIPLE_TAP;
  }
  else if(getBit(IQSMemoryMap.DOUBLE_FINGER_GESTURES[0], IQS9150_GESTURE_PRESS_HOLD_BIT))
  {
    return IQS9150_DOUBLE_GESTURE_PRESS_HOLD;
  }
  else if(getBit(IQSMemoryMap.DOUBLE_FINGER_GESTURES[0], IQS9150_DOUBLE_GESTURE_ZOOM_IN_BIT))
  {
    return IQS9150_DOUBLE_GESTURE_ZOOM_IN;
  }
  else if(getBit(IQSMemoryMap.DOUBLE_FINGER_GESTURES[0], IQS9150_DOUBLE_GESTURE_ZOOM_OUT_BIT))
  {
    return IQS9150_DOUBLE_GESTURE_ZOOM_OUT;
  }
  else if(getBit(IQSMemoryMap.DOUBLE_FINGER_GESTURES[0], IQS9150_DOUBLE_GESTURE_VERTICAL_SCROLL_BIT))
  {
    return IQS9150_DOUBLE_GESTURE_VERTICAL_SCROLL;
  }
  else if(getBit(IQSMemoryMap.DOUBLE_FINGER_GESTURES[0], IQS9150_DOUBLE_GESTURE_HORIZONTAL_SCROLL_BIT))
  {
    return IQS9150_DOUBLE_GESTURE_HORIZONTAL_SCROLL;
  }
  else
  {
    return IQS9150_GESTURE_NONE;
  }
}

/**
  * @name   getNumFingers
  * @brief  A method that returns the number of fingers active on the trackpad.
  * @param  None.
  * @retval Returns an 8-bit unsigned integer value of the number of fingers
  *         on the trackpad.
  */
uint8_t IQS9150::getNumFingers(void)
{

  uint8_t numFingers = IQSMemoryMap.TRACKPAD_FLAGS[0]; // load the system status byte
  numFingers = numFingers & 0x0F; // extract the power mode bits
  return numFingers;
}

/*****************************************************************************/
/*									     		ADVANCED PUBLIC METHODS							    	 		   */
/*****************************************************************************/


/*****************************************************************************/
/*                            PRIVATE METHODS                                */
/*****************************************************************************/
/**
 * @name    readRandomBytes8
 * @brief   A method that reads a specified number of bytes from a specified
 *          8 bit address and saves it into a user-supplied array.
 * @param   memoryAddress -> The memory address at which to start reading bytes
 *                           from.
 * @param   numBytes      -> The number of bytes that must be read.
 * @param   bytesArray    -> The array which will store the bytes to be read,
 *                           this array will be overwritten.
 * @param   stopOrRestart -> A boolean that specifies whether the communication
 *                           window should remain open or be closed after transfer.
 *                           False keeps it open, true closes it. Use the STOP
 *                           and RESTART definitions.
 * @retval  No value is returned, however, the user-supplied array is overwritten.
 * @note    Uses standard Arduino "Wire" library which is for I2C communication.
 *          Take note that C++ cannot return an array, therefore, the array which
 *          is passed as an argument is overwritten with the required values.
 *          Pass an array to the method by using only its name, e.g. "bytesArray",
 *          without the brackets, this passes a pointer to the array.
 */
void IQS9150::readRandomBytes8(uint8_t deviceAddress, uint8_t memoryAddress, uint8_t numBytes, uint8_t bytesArray[], bool stopOrRestart)
{
	uint8_t i = 0;  // A simple counter to assist with loading bytes into the user supplied array.
	uint8_t error_s = 0; // I2C Error variable to track any issues during communication

	/* Select the device with the address of "deviceAddress" and start communication. */
	Wire.beginTransmission(deviceAddress);
	/* Send a bit asking for the "memoryAddress" register. */
	Wire.write(memoryAddress);
	/* Complete the selection and communication initialization. */
	error_s = Wire.endTransmission(RESTART);  // Restart transmission for reading that follows.
	/* The required device has now been selected and it has been told which register to send information from. */

	/* Request "numBytes" bytes from the device which has address "deviceAddress"*/
	uint8_t counter = 0;
	do
	{
		Wire.requestFrom((int)deviceAddress, (int)numBytes, (int)stopOrRestart);

		/* break out of request loop if max retry is reached */
		if(counter++ >= IQS9150_I2C_RETRY)
		{
			return;
		}
	}while(Wire.available() == 0);  // Wait for response, this sometimes takes a few attempts

	/* Load the received bytes into the array until there are no more */
	while(Wire.available())
	{
		/* Load the received bytes into the user supplied array */
		bytesArray[i] = Wire.read();
		i++;
	}

  /* Always manually close the RDY window after a STOP is sent to prevent
     writing while the RDY window closes */
  if(stopOrRestart == STOP)
  {
    iqs9150_deviceRDY = false;
  }
}

/**
 * @name    readRandomBytes16
 * @brief   A method that reads a specified number of bytes from a specified
 *          16 bit address and saves it into a user-supplied array.
 * @param   memoryAddress -> The memory address at which to start reading bytes
 *                           from.
 * @param   numBytes      -> The number of bytes that must be read.
 * @param   bytesArray    -> The array which will store the bytes to be read,
 *                           this array will be overwritten.
 * @param   stopOrRestart -> A boolean that specifies whether the communication
 *                           window should remain open or be closed after transfer.
 *                           False keeps it open, true closes it. Use the STOP
 *                           and RESTART definitions.
 * @retval  No value is returned, however, the user-supplied array is overwritten.
 * @note    Uses standard Arduino "Wire" library which is for I2C communication.
 *          Take note that C++ cannot return an array, therefore, the array which
 *          is passed as an argument is overwritten with the required values.
 *          Pass an array to the method by using only its name, e.g. "bytesArray",
 *          without the brackets, this passes a pointer to the array.
 */
void IQS9150::readRandomBytes16(uint8_t deviceAddress, uint16_t memoryAddress, uint8_t numBytes, uint8_t bytesArray[], bool stopOrRestart)
{
	uint8_t i = 0;  // A simple counter to assist with loading bytes into the user supplied array.
	uint8_t error_s = 0; // I2C Error variable to track any issues during communication

	// Select the device with the address of "deviceAddress" and start communication.
	Wire.beginTransmission(deviceAddress);
	/* Specify the memory address */
	uint8_t addr_h, addr_l;
	addr_h = memoryAddress >> 8;
	addr_l = memoryAddress;
	Wire.write(addr_l);
	Wire.write(addr_h);

	/* Complete the selection and communication initialization. */
	error_s = Wire.endTransmission(RESTART);  // Restart transmission for reading that follows.
	/* The required device has now been selected and it has been told which register to send information from. */

	/* Request "numBytes" bytes from the device which has address "deviceAddress"*/
	uint8_t counter = 0;
	do
	{
		Wire.requestFrom((int)deviceAddress, (int)numBytes, (int)stopOrRestart);

		/* break out of request loop if max retry is reached */
		if(counter++ >= IQS9150_I2C_RETRY)
		{
			return;
		}
	}while(Wire.available() == 0);  // Wait for response, this sometimes takes a few attempts

	/* Load the received bytes into the array until there are no more */
	while(Wire.available())
	{
		/* Load the received bytes into the user supplied array */
		bytesArray[i] = Wire.read();
		i++;
	}

  /* Always manually close the RDY window after a STOP is sent to prevent
     writing while the RDY window closes */
  if(stopOrRestart == STOP)
  {
    iqs9150_deviceRDY = false;
  }
}

/**
  * @name   writeRandomBytes8
  * @brief  A method that writes a specified number of bytes to a specified
  *         8 bit address, the bytes to write are supplied using an array pointer.
  * @param  deviceAddress -> The slave device address to which the communication should be transmitted.
  * 		memoryAddress -> The memory address at which to start writing the
  *         bytes to.
  *         numBytes      -> The number of bytes that must be written.
  *         bytesArray    -> The array which stores the bytes which will be
  *                          written to the memory location.
  *         stopOrRestart -> A boolean that specifies whether the communication
  *                          window should remain open or be closed of transfer.
  *                          False keeps it open, true closes it. Use the STOP
  *                          and RESTART definitions.
  * @retval No value is returned, only the IQS device registers are altered.
  * @note   Uses standard Arduino "Wire" library which is for I2C communication.
  *         Take note that a full array cannot be passed to a function in C++.
  *         Pass an array to the function by using only its name, e.g. "bytesArray",
  *         without the square brackets, this passes a pointer to the
  *         array. The values to be written must be loaded into the array prior
  *         to passing it to the function.
  */
void IQS9150::writeRandomBytes8(uint8_t deviceAddress, uint8_t memoryAddress, uint8_t numBytes, uint8_t bytesArray[], bool stopOrRestart)
{
	uint8_t error_s = 0; // I2C Error variable to track any issues during communication

	/* Select the device with the address of "deviceAddress" and start communication. */
	Wire.beginTransmission(deviceAddress);
	/* Specify the memory address */
	Wire.write(memoryAddress);
	/* Write the bytes as specified in the array which "arrayAddress" pointer points to. */
	for(int i=0; i<numBytes; i++)
	{
		Wire.write(bytesArray[i]);
	}
	/* End the transmission, user decides to STOP or RESTART. */
	error_s = Wire.endTransmission(stopOrRestart);
}

/**
  * @name   writeRandomBytes16
  * @brief  A method that writes a specified number of bytes to a specified
  *         16 bit address, the bytes to write are supplied using an array pointer.
  * @param  deviceAddress -> The slave device address to which the communication should be transmitted.
  * 		memoryAddress -> The memory address at which to start writing the
  *         bytes to.
  *         numBytes      -> The number of bytes that must be written.
  *         bytesArray    -> The array which stores the bytes which will be
  *                          written to the memory location.
  *         stopOrRestart -> A boolean that specifies whether the communication
  *                          window should remain open or be closed of transfer.
  *                          False keeps it open, true closes it. Use the STOP
  *                          and RESTART definitions.
  * @retval No value is returned, only the IQS device registers are altered.
  * @note   Uses standard Arduino "Wire" library which is for I2C communication.
  *         Take note that a full array cannot be passed to a function in C++.
  *         Pass an array to the function by using only its name, e.g. "bytesArray",
  *         without the square brackets, this passes a pointer to the
  *         array. The values to be written must be loaded into the array prior
  *         to passing it to the function.
  */
void IQS9150::writeRandomBytes16(uint8_t deviceAddress, uint16_t memoryAddress, uint8_t numBytes, uint8_t bytesArray[], bool stopOrRestart)
{
	uint8_t error_s = 0; // I2C Error variable to track any issues during communication

	// Select the device with the address of "deviceAddress" and start communication.
	Wire.beginTransmission(deviceAddress);

	/* Specify the memory address */
	uint8_t addr_h, addr_l;
	addr_h = memoryAddress >> 8;
	addr_l = memoryAddress;
	Wire.write(addr_l);
	Wire.write(addr_h);

	// Write the bytes as specified in the array which "arrayAddress" pointer points to.
	for(int i=0; i<numBytes; i++)
	{
	Wire.write(bytesArray[i]);
	}
	// End the transmission, user decides to STOP or RESTART.
	error_s = Wire.endTransmission(stopOrRestart);

  /* Always manually close the RDY window after a STOP is sent to prevent
     writing while the RDY window closes */
  if(stopOrRestart == STOP)
  {
    iqs9150_deviceRDY = false;
  }
}

/**
  * @name   getBit
  * @brief  A method that returns the chosen bit value of the provided byte.
  * @param  data       -> byte of which a given bit value needs to be calculated.
  * @param  bit_number -> a number between 0 and 7 representing the bit in question.
  * @retval The boolean value of the specific bit requested.
  */
bool IQS9150::getBit(uint8_t data, uint8_t bit_number)
{
  return (data & ( 1 << bit_number )) >> bit_number;
}

/**
  * @name   setBit
  * @brief  A method that returns the chosen bit value of the provided byte.
  * @param  data       -> byte of which a given bit value needs to be calculated.
  * @param  bit_number -> a number between 0 and 7 representing the bit in question.
  * @retval Returns an 8-bit unsigned integer value of the given data byte with
  *         the requested bit set.
  */
uint8_t IQS9150::setBit(uint8_t data, uint8_t bit_number)
{
	return (data |= 1UL << bit_number);
}

/**
  * @name   clearBit
  * @brief  A method that returns the chosen bit value of the provided byte.
  * @param  data       -> byte of which a given bit value needs to be calculated.
  * @param  bit_number -> a number between 0 and 7 representing the bit in question.
  * @retval Returns an 8-bit unsigned integer value of the given data byte with
  *         the requested bit cleared.
  */
uint8_t IQS9150::clearBit(uint8_t data, uint8_t bit_number)
{
	return (data &= ~(1UL << bit_number));
}

/**
  * @name   force_I2C_communication
  * @brief  A method which writes data 0x00 to memory address 0xFF to open a
  *         communication window on the IQS9150.
  * @param  None.
  * @retval None.
  * @note   Uses standard Arduino "Wire" library which is for I2C communication.
  */
void IQS9150::force_I2C_communication(void)
{
  /*Ensure RDY is HIGH at the moment*/
  if (!iqs9150_deviceRDY)
  {
    /* Select the device with the address: "DEMO_IQS9150_ADDR" and start
      communication. */
    Wire.beginTransmission(_deviceAddress);

    /* Write to memory address 0xFF that will prompt the IQS9150 to open a
    communication window.*/
    Wire.write(0xFF);
    Wire.write(0x00);

    /* End the transmission, user decides to STOP or RESTART. */
    Wire.endTransmission(STOP);
  }
}
