/******************************************************************************
 *                                                                            *
 *                                                                            *
 *                                Copyright by                                *
 *                                                                            *
 *                              Azoteq (Pty) Ltd                              *
 *                          Republic of South Africa                          *
 *                                                                            *
 *                           Tel: +27(0)21 863 0033                           *
 *                           E-mail: info@azoteq.com                          *
 *                                                                            *
 * ========================================================================== *
 * @file        IQS9150.h                                                     *
 * @brief       This file contains all the necessary settings for the IQS9150 *
 *              and this file can be changed from the GUI or edited here      *
 * @author      CC van Eeden - Azoteq PTY Ltd                                 *
 * @version     v1.0                                                          *
 * @date        2024                                                          *
 * ========================================================================== *
 * @attention  Makes use of the following standard Arduino libraries:         *
 * - Arduino.h -> Included in IQS9150.h, comes standard with Arduino          *
 * - Wire.h    -> Included in IQS9150.h, comes standard with Arduino          *
 ******************************************************************************/

#ifndef IQS9150_H
#define IQS9150_H

// Include Files
#include "Arduino.h"
#include "Wire.h"
#include "./inc/iqs9150_addresses.h"

#define IQS9150_I2C_RETRY               10

// Public Global Definitions
/* For use with Wire.h library. True argument with some functions closes the
   I2C communication window.*/
#define STOP    true
/* For use with Wire.h library. False argument with some functions keeps the
   I2C communication window open. */
#define RESTART false

//For use with certain function that apply to either TP or ALP separately
#define TP      true
#define ALP     false

// Device Info
#define IQS9150_PRODUCT_NUM                             1898

// System event bits
#define IQS9150_TP_RE_ATI_OCCURRED_BIT                  4
#define IQS9150_ALP_RE_ATI_OCCURRED_BIT                 6
#define IQS9150_SHOW_RESET_BIT	                        7

// Utility Bits
#define IQS9150_ACK_RESET_BIT		                7
#define IQS9150_SW_RESET_BIT		                1
#define IQS9150_TP_RE_ATI_BIT		                5
#define IQS9150_ALP_RE_ATI_BIT		                6
#define IQS9150_TP_RESEED_BIT		                3
#define IQS9150_ALP_RESEED_BIT		                4


// Config Settings Bits
#define IQS9150_EVENT_MODE_BIT	                        0       // 8

// Gesture Bits
#define IQS9150_GESTURE_SINGLE_TAP_BIT                           0
#define IQS9150_GESTURE_DOUBLE_TAP_BIT                           1
#define IQS9150_GESTURE_TRIPLE_TAP_BIT                           2
#define IQS9150_GESTURE_PRESS_HOLD_BIT                           3
#define IQS9150_SINGLE_GESTURE_PALM_GESTURE_BIT                  4
#define IQS9150_SINGLE_GESTURE_SWIPE_X_POSITIVE_BIT              0 // 8
#define IQS9150_SINGLE_GESTURE_SWIPE_X_NEGATIVE_BIT              1 // 9
#define IQS9150_SINGLE_GESTURE_SWIPE_Y_POSITIVE_BIT              2 // 10
#define IQS9150_SINGLE_GESTURE_SWIPE_Y_NEGATIVE_BIT              3 // 11
#define IQS9150_SINGLE_GESTURE_SWIPE_HOLD_X_POSITIVE_BIT         4 // 12
#define IQS9150_SINGLE_GESTURE_SWIPE_HOLD_X_NEGATIVE_BIT         5 // 13
#define IQS9150_SINGLE_GESTURE_SWIPE_HOLD_Y_POSITIVE_BIT         6 // 14
#define IQS9150_SINGLE_GESTURE_SWIPE_HOLD_Y_NEGATIVE_BIT         7 // 15
#define IQS9150_DOUBLE_GESTURE_ZOOM_IN_BIT                       4
#define IQS9150_DOUBLE_GESTURE_ZOOM_OUT_BIT                      5
#define IQS9150_DOUBLE_GESTURE_VERTICAL_SCROLL_BIT               6
#define IQS9150_DOUBLE_GESTURE_HORIZONTAL_SCROLL_BIT             7

/* Defines and structs for IQS9150 states */
/**
* @brief  iqs9150 Init Enumeration.
*/
typedef enum {
        IQS9150_INIT_NONE = (uint8_t) 0x00,
        IQS9150_INIT_VERIFY_PRODUCT,
        IQS9150_INIT_READ_RESET,
	IQS9150_INIT_CHIP_RESET,
        IQS9150_INIT_ACK_RESET,
        IQS9150_INIT_CHECK_RESET,
        IQS9150_INIT_UPDATE_SETTINGS,
        IQS9150_INIT_ATI,
        IQS9150_INIT_WAIT_FOR_ATI,
        IQS9150_INIT_READ_DATA,
        IQS9150_INIT_RESEED,
	IQS9150_INIT_ACTIVATE_EVENT_MODE,
        IQS9150_INIT_ACTIVATE_STREAM_MODE,
	IQS9150_INIT_DONE
} iqs9150_init_e;

typedef enum {
        IQS9150_STATE_NONE = (uint8_t) 0x00,
        IQS9150_STATE_START,
        IQS9150_STATE_INIT,
        IQS9150_STATE_SW_RESET,
        IQS9150_STATE_CHECK_RESET,
	IQS9150_STATE_RUN,
} iqs9150_state_e;

typedef enum {
        IQS9150_ACTIVE_MODE = (uint8_t) 0x00,
        IQS9150_IDLE_TOUCH_MODE,
        IQS9150_IDLE_MODE,
        IQS9150_LP1_MODE,
        IQS9150_LP2_MODE,
        IQS9150_POWER_UNKNOWN,
} iqs9150_power_mode_e;

typedef enum {
        IQS9150_SINGLE_GESTURE_SINGLE_TAP = (uint8_t) 0x00,
        IQS9150_SINGLE_GESTURE_DOUBLE_TAP,
        IQS9150_SINGLE_GESTURE_TRIPLE_TAP,
        IQS9150_SINGLE_GESTURE_PRESS_HOLD,
        IQS9150_SINGLE_GESTURE_PALM_GESTURE,
        IQS9150_SINGLE_GESTURE_SWIPE_X_POSITIVE,
        IQS9150_SINGLE_GESTURE_SWIPE_X_NEGATIVE,
        IQS9150_SINGLE_GESTURE_SWIPE_Y_POSITIVE,
        IQS9150_SINGLE_GESTURE_SWIPE_Y_NEGATIVE,
        IQS9150_SINGLE_GESTURE_SWIPE_HOLD_X_POSITIVE,
        IQS9150_SINGLE_GESTURE_SWIPE_HOLD_X_NEGATIVE,
        IQS9150_SINGLE_GESTURE_SWIPE_HOLD_Y_POSITIVE,
        IQS9150_SINGLE_GESTURE_SWIPE_HOLD_Y_NEGATIVE,
        IQS9150_DOUBLE_GESTURE_SINGLE_TAP,
        IQS9150_DOUBLE_GESTURE_DOUBLE_TAP,
        IQS9150_DOUBLE_GESTURE_TRIPLE_TAP,
        IQS9150_DOUBLE_GESTURE_PRESS_HOLD,
        IQS9150_DOUBLE_GESTURE_ZOOM_IN,
        IQS9150_DOUBLE_GESTURE_ZOOM_OUT,
        IQS9150_DOUBLE_GESTURE_VERTICAL_SCROLL,
        IQS9150_DOUBLE_GESTURE_HORIZONTAL_SCROLL,
        IQS9150_GESTURE_NONE
} iqs9150_gestures_e;

/* IQS9150 Memory map data variables, only save the data that might be used
during program runtime */
#pragma pack(1)
typedef struct
{
	/* READ ONLY */			        //  I2C Addresses:
	uint8_t SINGLE_FINGER_GESTURES[2]; 	// 	0x101C
	uint8_t DOUBLE_FINGER_GESTURES[2];      // 	0x101E
        uint8_t INFO_FLAGS[2];                  //      0x1020
        uint8_t TRACKPAD_FLAGS[2];              //      0x1022

        uint8_t FINGER_1_X[2];                  //      0x1024
        uint8_t FINGER_1_Y[2];                  //      0x1026
        uint8_t FINGER_1_TOUCH_STRENGTH[2];     //      0x1028
        uint8_t FINGER_1_AREA[2];               //      0x102A

        uint8_t FINGER_2_X[2];                  //      0x102C
        uint8_t FINGER_2_Y[2];                  //      0x102E
        uint8_t FINGER_2_TOUCH_STRENGTH[2];     //      0x1030
        uint8_t FINGER_2_AREA[2];               //      0x1032

        uint8_t FINGER_3_X[2];                  //      0x1034
        uint8_t FINGER_3_Y[2];                  //      0x1036
        uint8_t FINGER_3_TOUCH_STRENGTH[2];     //      0x1038
        uint8_t FINGER_3_AREA[2];               //      0x103A

        uint8_t FINGER_4_X[2];                  //      0x103C
        uint8_t FINGER_4_Y[2];                  //      0x103E
        uint8_t FINGER_4_TOUCH_STRENGTH[2];     //      0x1040
        uint8_t FINGER_4_AREA[2];               //      0x1042

        uint8_t FINGER_5_X[2];                  //      0x1044
        uint8_t FINGER_5_Y[2];                  //      0x1046
        uint8_t FINGER_5_TOUCH_STRENGTH[2];     //      0x1048
        uint8_t FINGER_5_AREA[2];               //      0x104A

        uint8_t FINGER_6_X[2];                  //      0x104C
        uint8_t FINGER_6_Y[2];                  //      0x104E
        uint8_t FINGER_6_TOUCH_STRENGTH[2];     //      0x1050
        uint8_t FINGER_6_AREA[2];               //      0x1052

        uint8_t FINGER_7_X[2];                  //      0x1054
        uint8_t FINGER_7_Y[2];                  //      0x1056
        uint8_t FINGER_7_TOUCH_STRENGTH[2];     //      0x1058
        uint8_t FINGER_7_AREA[2];               //      0x105A

	/* READ WRITE */		        //  I2C Addresses:
	uint8_t SYSTEM_CONTROL[2]; 	        // 	0x11BC
} IQS9150_MEMORY_MAP;
#pragma pack(4)

#pragma pack(1)
typedef struct {
        iqs9150_state_e        state;
        iqs9150_init_e         init_state;
}iqs9150_s;
#pragma pack(4)

// Class Prototype
class IQS9150
{
public:
        // Public Constructors
        IQS9150();

        // Public Device States
        iqs9150_s iqs9150_state;

        // Public Variables
        IQS9150_MEMORY_MAP IQSMemoryMap;
        bool new_data_available;

        // Public Methods
        void begin(uint8_t deviceAddressIn, uint8_t rdy_pin);
        bool init(void);
        void run(void);
        void queueValueUpdates(void);
        void requestData(void);

        uint16_t getProductNum(bool stopOrRestart);
        uint8_t getmajorVersion(bool stopOrRestart);
        uint8_t getminorVersion(bool stopOrRestart);


        void updateInfoFlags(bool stopOrRestart);
        bool checkReset(void);

        void updateSettings(bool stopOrRestart);
        iqs9150_power_mode_e getPowerMode(void);
        iqs9150_gestures_e getTouchpadEvent(void);
        uint8_t getNumFingers(void);
        uint16_t getAbsXCoordinate(uint8_t fingerNum);
        uint16_t getAbsYCoordinate(uint8_t fingerNum);

        void force_I2C_communication(void);

private:
        // Private Variables
        uint8_t _deviceAddress;

        // Private Methods
        void readRandomBytes8(uint8_t deviceAddress, uint8_t memoryAddress, uint8_t numBytes, uint8_t bytesArray[], bool stopOrRestart);
        void readRandomBytes16(uint8_t deviceAddress, uint16_t memoryAddress, uint8_t numBytes, uint8_t bytesArray[], bool stopOrRestart);
        void writeRandomBytes8(uint8_t deviceAddress, uint8_t memoryAddress, uint8_t numBytes, uint8_t bytesArray[], bool stopOrRestart);
        void writeRandomBytes16(uint8_t deviceAddress, uint16_t memoryAddress, uint8_t numBytes, uint8_t bytesArray[], bool stopOrRestart);
        bool getBit(uint8_t data, uint8_t bit_number);
        uint8_t setBit(uint8_t data, uint8_t bit_number);
        uint8_t clearBit(uint8_t data, uint8_t bit_number);
        void clearRDY(void);
        bool getRDYStatus(void);

        bool readATIoccurred(bool tpOrAlp = TP);
        void acknowledgeReset(bool stopOrRestart);
        void ReATI(bool tpOrAlp = TP);
        void ReSeed(bool tpOrAlp = TP);
        void setEventMode(void);
        void SW_Reset(bool stopOrRestart);
};
#endif // IQS9150_h
